__docformat__ = 'restructuredtext'

from petsc4py import PETSc

from fipy.tools.timer import Timer
from .petscSolver import PETScSolver
from .preconditioners.petscPreconditioner import PETScPreconditioner
from .preconditioners.defaultPreconditioner import DefaultPreconditioner

from fipy.tools import numerix

__all__ = ["PETScKrylovSolver"]

class PETScKrylovSolver(PETScSolver):

    """
    .. attention:: This class is abstract, always create one of its subclasses.
       It provides the code to call all Krylov solvers from the PETSc package.

    """

    DEFAULT_PRECONDITIONER = DefaultPreconditioner

    def __init__(self, tolerance="default",
                 absolute_tolerance=None,
                 divergence_tolerance=None,
                 criterion="default",
                 iterations="default", precon="default"):
        """
        Parameters
        ----------
        tolerance : float
            Required relative error tolerance.
        absolute_tolerance : float
            Required absolute error tolerance.
        divergence_tolerance : float
            Required divergence error tolerance.
        criterion : {'default', 'unscaled', 'RHS', 'matrix', 'initial', 'preconditioned', 'natural', 'legacy'}
            Interpretation of ``tolerance``.
            See :ref:`CONVERGENCE` for more information.
        iterations : int
            Maximum number of iterative steps to perform.
        precon : ~fipy.solvers.petsc.preconditioners.petscPreconditioner.PETScPreconditioner or str, optional
            Preconditioner to apply to the matrix.  A value of None means
            to perform an unpreconditioned solve.  While, for consistency
            with other solver suites, :term:`FiPy` does implement some
            preconditioner objects for PETSc, you can also simply pass one
            of the `PCType strings`_.  (default:
            :class:`~fipy.solvers.petsc.preconditioners.defaultPreconditioner.DefaultPreconditioner`).

            .. _PCType strings: https://www.mcs.anl.gov/petsc/petsc-current/docs/manualpages/PC/PCType.html
        """
        if self.__class__ is PETScKrylovSolver:
            raise NotImplementedError("can't instantiate abstract base class")

        self.absolute_tolerance = absolute_tolerance
        self.divergence_tolerance = divergence_tolerance
        super(PETScKrylovSolver, self).__init__(tolerance=tolerance, criterion=criterion,
                                                iterations=iterations, precon=precon)

    def _adaptLegacyTolerance(self, L, x, b):
        return (1., PETSc.KSP.NormType.DEFAULT)

    def _convergenceTest(self, ksp, its, rnorm, scale):
        """Replace `KSPConvergedDefault()` with custom normalization

        Modeled on
        `KSPConvergedDefault() <https://gitlab.com/petsc/petsc/-/blob/main/src/ksp/ksp/interface/iterativ.c#L1512>`_.
        Simplistically, converged if `rnorm <= rtol * scale`.

        It would be much nicer (and less expensive) if they would just let
        you specify how to calculate `rnorm0`!
        `KSPConvergedDefaultSetUIRNorm()` isn't exposed to petsc4py, and it
        wouldn't help with "unscaled" or "matrix".

        Parameters
        ----------
        ksp : ~petsc4py.PETSc.KSP
            Krylov solver object.
        its : int
            Number of iterations performed.
        rnorm : float
            Norm of the residual to test.
        scale : float
            How to interpret magnitude of tolerance.

        Returns
        -------
        bool
            Whether solution is converged.
        """

        reason = PETSc.KSP.ConvergedReason.CONVERGED_ITERATING

        min_it = 0

        if numerix.isnan(rnorm) or numerix.isinf(rnorm):
            pcreason = PCReduceFailedReason
            if pcreason:
                reason = PETSc.KSP.ConvergedReason.DIVERGED_PCSETUP_FAILED
                self._log.debug("Linear solver pcsetup fails, "
                                "declaring divergence")
            else:
                reason = PETSc.KSP.ConvergedReason.DIVERGED_NANORINF
                self._log.debug("Linear solver has created a not a number (NaN) "
                                "as the residual norm, declaring divergence")
        elif its >= min_it:
            if rnorm <= max(ksp.rtol * scale, ksp.atol):
                if rnorm < ksp.atol:
                    reason = PETSc.KSP.ConvergedReason.CONVERGED_ATOL
                    self._log.debug("Linear solver has converged. "
                                    "Residual norm {rnorm:14.12e} is less "
                                    "than absolute tolerance {atol:14.12e} "
                                    "at iteration {its}".format(rnorm=rnorm,
                                                                atol=ksp.atol,
                                                                its=its))
                else:
                    reason = PETSc.KSP.ConvergedReason.CONVERGED_RTOL
                    self._log.debug("Linear solver has converged. "
                                    "Residual norm {rnorm:14.12e} is less "
                                    "than relative tolerance {rtol:14.12e} "
                                    "times residual scale {scale:14.12e} "
                                    "at iteration {its}".format(rnorm=rnorm,
                                                                rtol=ksp.rtol,
                                                                scale=scale,
                                                                its=its))
        elif rnorm >= ksp.dtol * scale:
            reason = PETSc.KSP.ConvergedReason.DIVERGED_DTOL
            self._log.debug("Linear solver is diverging. "
                            "Residual scale {scale:14.12e}, "
                            "current residual norm {rnorm:14.12e} "
                            "at iteration {its}".format(rnorm=rnorm,
                                                        scale=scale,
                                                        its=its))

        return reason

    def _adaptUnscaledTolerance(self, L, x, b):
        return (1., self._convergenceTest)

    def _adaptRHSTolerance(self, L, x, b):
        return (1., PETSc.KSP.NormType.UNPRECONDITIONED)

    def _adaptMatrixTolerance(self, L, x, b):
        return (self._matrixNorm(L, x, b), self._convergenceTest)

    def _adaptInitialTolerance(self, L, x, b):
        return (self._residualNorm(L, x, b), self._convergenceTest)

    def _adaptPreconditionedTolerance(self, L, x, b):
        return (1., PETSc.KSP.NormType.PRECONDITIONED)

    def _adaptNaturalTolerance(self, L, x, b):
        return (1., PETSc.KSP.NormType.NATURAL)

    def _solve_(self, L, x, b):
        """Solve system of equations posed for PETSc

        Parameters
        ----------
        L : PETSc.Mat
            Sparse matrix
        x : PETSc.Vec
            Solution variable as ghosted vector
        b : PETSc.Vec
            Right-hand side as ghosted vector

        Returns
        -------
        x : PETSc.Vec
            Solution variable as ghosted vector
        """
        ksp = PETSc.KSP()
        ksp.create(L.comm)
        ksp.setType(self.solver)
        if self.criterion != "legacy":
            ksp.setInitialGuessNonzero(True)

        L.assemble()
        ksp.setOperators(L)

        tolerance_scale, suite_criterion = self._adaptTolerance(L, x, b)
        if suite_criterion == self._convergenceTest:
            ksp.setConvergenceTest(suite_criterion,
                                   kargs=dict(scale=tolerance_scale))
            suite_criterion = PETSc.KSP.NormType.PRECONDITIONED
            rtol = self.tolerance
            divtol = self.divergence_tolerance
        else:
            rtol, divtol = (self.scale_tolerance(tol, tolerance_scale)
                            for tol in (self.tolerance,
                                        self.divergence_tolerance))

        ksp.setTolerances(rtol=rtol,
                          atol=self.absolute_tolerance,
                          divtol=divtol,
                          max_it=self.iterations)
        ksp.setNormType(suite_criterion)

        self._log.debug("BEGIN precondition")

        with Timer() as t:
            if self.preconditioner is None:
                ksp.getPC().setType("none")
            elif isinstance(self.preconditioner, PETScPreconditioner):
                self.preconditioner._applyToSolver(solver=ksp, matrix=L)
            else:
                ksp.getPC().setType(self.preconditioner)

        self._log.debug("END precondition - {} ns".format(t.elapsed))

        ksp.setFromOptions()

        self._log.debug("BEGIN solve")

        with Timer() as t:
            ksp.solve(b, x)

        self._log.debug("END solve - {} ns".format(t.elapsed))

        self._setConvergence(suite="petsc",
                             code=ksp.reason,
                             iterations=ksp.its,
                             tolerance_scale=tolerance_scale,
                             # "The residual value that is tested may be an approximation"
                             # https://petsc.org/release/petsc4py/reference/petsc4py.PETSc.KSP.html#petsc4py.PETSc.KSP.setConvergenceTest
                             residual=ksp.norm,
                             ksp_solver=ksp.type,
                             ksp_precon=ksp.getPC().type,
                             ksp_norm_type=ksp.norm_type)

        ksp.destroy()

        self.convergence.warn()

        return x



r"""
A simple 1D example to test the setup of the Poisson equation.

>>> from fipy import CellVariable, Grid1D, DiffusionTerm, Viewer
>>> from fipy.tools import numerix

>>> nx = 200
>>> dx = 0.01
>>> L = nx * dx
>>> mesh = Grid1D(dx = dx, nx = nx)

The dimensionless Poisson equation is

.. math::

   \nabla\cdot\left(\epsilon\nabla\phi\right) = -\rho = -\sum_{j=1}^n z_j C_j

where :math:`\phi` is the electrostatic potential, :math:`\epsilon` is the
permittivity, :math:`\rho` is the charge density, :math:`C_j` is the
concentration of the :math:`j^\text{th}` component, and :math:`z_j` is the
valence of the :math:`j^\text{th}` component.

We will be solving for the electrostatic potential

>>> potential = CellVariable(mesh = mesh, name = 'phi', value = 0.)
>>> permittivity = 1.

We examine a fixed distribution of electrons with :math:`z_{\text{e}^{-}} = -1`.

>>> class ComponentVariable(CellVariable):
...     def __init__(self, mesh, value = 0., name = '',
...                  standardPotential = 0., barrier = 0.,
...                  diffusivity = None, valence = 0, equation = None):
...         CellVariable.__init__(self, mesh = mesh,
...                               value = value, name = name)
...         self.standardPotential = standardPotential
...         self.barrier = barrier
...         self.diffusivity = diffusivity
...         self.valence = valence
...         self.equation = equation
... 
...     def copy(self):
...         return self.__class__(mesh = self.mesh,
...                               value = self.value,
...                               name = self.name,
...                               standardPotential =
...                                   self.standardPotential,
...                               barrier = self.barrier,
...                               diffusivity = self.diffusivity,
...                               valence = self.valence,
...                               equation = self.equation)

Since we're only interested in a single species, electrons, we could
simplify the following, but since we will in general be studying multiple
components, we explicitly allow for multiple substitutional species and
multiple interstitial species:

>>> interstitials = [
...     ComponentVariable(mesh = mesh, name = 'e-', valence = -1)]
>>> substitutionals = []

Because Poisson's equation admits an infinite number of potential profiles,
we must constrain the solution by fixing the potential at one point:

>>> potential.constrain(0., mesh.facesLeft)

>>> charge = 0.
>>> for Cj in interstitials + substitutionals:
...     charge += Cj * Cj.valence

>>> potential.equation = DiffusionTerm(coeff = permittivity) \
...                      + charge == 0

First, we obtain a uniform charge distribution by setting a uniform
concentration of electrons :math:`C_{\text{e}^{-}} = 1`.

>>> interstitials[0].setValue(1.)

and we solve for the electrostatic potential

>>> potential.equation.solve(var = potential)

This problem has the analytical solution

.. math::

   \psi(x) = \frac{x^2}{2} - 2x

We verify that the correct equilibrium is attained

>>> x = mesh.cellCenters[0]
>>> analyticalArray = (x**2)/2 - 2*x

>>> print(potential.allclose(analyticalArray, rtol = 2e-5, atol = 2e-5))
1

If we are running the example interactively, we view the result

>>> from fipy import input
>>> if __name__ == '__main__':
...     viewer = Viewer(vars = (charge, potential))
...     viewer.plot()
...     input("Press any key to continue...")

Next, we segregate all of the electrons to right side of the domain

.. math::

   C_{\text{e}^{-}} =
   \begin{cases}
       0& \text{for $x \le L/2$,} \\
       1& \text{for $x > L/2$.}
   \end{cases}

>>> x = mesh.cellCenters[0]
>>> interstitials[0].setValue(0.)
>>> interstitials[0].setValue(1., where=x > L / 2.)

and again solve for the electrostatic potential

>>> potential.equation.solve(var = potential)

which now has the analytical solution

.. math::

   \psi(x) =
   \begin{cases}
       -x& \text{for $x \le L/2$,} \\
       \frac{(x-1)^2}{2} - x& \text{for $x > L/2$.}
   \end{cases}

We verify that the correct equilibrium is attained

>>> analyticalArray = numerix.where(x < L/2, -x, ((x-1)**2)/2 - x)

>>> potential.allclose(analyticalArray, rtol = 2e-5, atol = 2e-5).value
1

and again view the result

>>> from fipy import input
>>> if __name__ == '__main__':
...     viewer.plot()
...     input("Press any key to continue...")

Finally, we segregate all of the electrons to left side of the domain

.. math::

   C_{\text{e}^{-}} =
   \begin{cases}
       1& \text{for $x \le L/2$,} \\
       0& \text{for $x > L/2$.}
   \end{cases}

>>> interstitials[0].setValue(1.)
>>> interstitials[0].setValue(0., where=x > L / 2.)

and again solve for the electrostatic potential

>>> potential.equation.solve(var = potential)

which has the analytical solution

.. math::

   \psi(x) =
   \begin{cases}
       \frac{x^2}{2} - x& \text{for $x \le L/2$,} \\
       -\frac{1}{2}& \text{for $x > L/2$.}
   \end{cases}

We again verify that the correct equilibrium is attained

>>> analyticalArray = numerix.where(x < 1, (x**2)/2 - x, -0.5)

>>> potential.allclose(analyticalArray, rtol = 2e-5, atol = 2e-5).value
1

and again view the result

>>> if __name__ == '__main__':
...     viewer.plot()
"""
__docformat__ = 'restructuredtext'

from fipy import input

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())

    input("finished")



r"""Solve a one-dimensional diffusion equation under different conditions.

To run this example from the base :term:`FiPy` directory, type::

    $ python examples/diffusion/mesh1D.py

at the command line. Different stages of the example should be displayed,
along with prompting messages in the terminal.

This example takes the user through assembling a simple problem with :term:`FiPy`.
It describes different approaches to a 1D diffusion problem with constant
diffusivity and fixed value boundary conditions such that,

.. math::
   :label: eq:diffusion:mesh1D:constantD

   \frac{\partial \phi}{\partial t} = D \nabla^2 \phi.

The first step is to define a one dimensional domain with 50 solution
points. The :class:`~fipy.meshes.grid1D.Grid1D` object represents a linear structured grid. The
parameter ``dx`` refers to the grid spacing (set to unity here).

>>> from fipy import Variable, FaceVariable, CellVariable, Grid1D, ExplicitDiffusionTerm, TransientTerm, DiffusionTerm, Viewer
>>> from fipy.tools import numerix

>>> nx = 50
>>> dx = 1.
>>> mesh = Grid1D(nx=nx, dx=dx)

:term:`FiPy` solves all equations at the centers of the cells of the mesh. We
thus need a :class:`~fipy.variables.cellVariable.CellVariable` object to hold the values of the
solution, with the initial condition :math:`\phi = 0` at :math:`t = 0`,

>>> phi = CellVariable(name="solution variable",
...                    mesh=mesh,
...                    value=0.)

We'll let

>>> D = 1.

for now.

The boundary conditions

.. math::

   \phi =
   \begin{cases}
       0& \text{at \(x = 1\),} \\
       1& \text{at \(x = 0\).}
   \end{cases}

are formed with a value

>>> valueLeft = 1
>>> valueRight = 0

and a set of faces over which they apply.

.. note::

   Only faces around the exterior of the mesh can be used for boundary
   conditions.

For example, here the exterior faces on the left of the domain are extracted by
``mesh``.\ :attr:`~fipy.meshes.abstractMesh.AbstractMesh.facesLeft`. The boundary
conditions are applied using
``phi``\. :meth:`~fipy.variables.variable.Variable.constrain` with these faces and
a value (``valueLeft``).

>>> phi.constrain(valueRight, mesh.facesRight)
>>> phi.constrain(valueLeft, mesh.facesLeft)

.. note::

   If no boundary conditions are specified on exterior faces, the default
   boundary condition is no-flux,
   :math:`\vec{n} \cdot (D \nabla \phi) \rvert_\text{someFaces} = 0`.

If you have ever tried to numerically solve
Eq. :eq:`eq:diffusion:mesh1D:constantD`,
you most likely attempted "explicit finite differencing" with code
something like::

    for step in range(steps):
        for j in range(cells):
            phi_new[j] = phi_old[j] \
              + (D * dt / dx**2) * (phi_old[j+1] - 2 * phi_old[j] + phi_old[j-1])
        time += dt

plus additional code for the boundary conditions. In :term:`FiPy`, you would write

.. index::
   single: ExplicitDiffusionTerm
   single: TransientTerm

>>> eqX = TransientTerm() == ExplicitDiffusionTerm(coeff=D)

The largest stable timestep that can be taken for this explicit 1D
diffusion problem is :math:`\Delta t \le \Delta x^2 / (2 D)`.

We limit our steps to 90% of that value for good measure

>>> timeStepDuration = 0.9 * dx**2 / (2 * D)
>>> steps = 100

If we're running interactively, we'll want to view the result, but not if
this example is being run automatically as a test. We accomplish this by
having Python check if this script is the "``__main__``" script, which will
only be true if we explicitly launched it and not if it has been imported
by another script such as the automatic tester. The factory function
:func:`~fipy.viewers.Viewer` returns a suitable viewer depending on available
viewers and the dimension of the mesh.

.. index::
   pair: module; fipy.viewers

>>> phiAnalytical = CellVariable(name="analytical value",
...                              mesh=mesh)

>>> if __name__ == '__main__':
...     viewer = Viewer(vars=(phi, phiAnalytical),
...                     datamin=0., datamax=1.)
...     viewer.plot()

In a semi-infinite domain, the analytical solution for this transient
diffusion problem is given by
:math:`\phi = 1 - \erf(x/2\sqrt{D t})`. If the :term:`SciPy` library is available,
the result is tested against the expected profile:

>>> x = mesh.cellCenters[0]
>>> t = timeStepDuration * steps

>>> try:
...     from scipy.special import erf # doctest: +SCIPY
...     phiAnalytical.setValue(1 - erf(x / (2 * numerix.sqrt(D * t)))) # doctest: +SCIPY
... except ImportError:
...     print("The SciPy library is not available to test the solution to \
... the transient diffusion equation")

We then solve the equation by repeatedly looping in time:

>>> for step in range(steps):
...     eqX.solve(var=phi,
...               dt=timeStepDuration)
...     if __name__ == '__main__':
...         viewer.plot()

>>> print(phi.allclose(phiAnalytical, atol = 7e-4)) # doctest: +SCIPY
1

>>> from fipy import input
>>> if __name__ == '__main__':
...     input("Explicit transient diffusion. Press <return> to proceed...")

.. image:: /figures/examples/diffusion/mesh1Dexplicit.*
   :width: 90%
   :align: center
   :alt: solution to diffusion problem evolved by explicit time steps

----

Although explicit finite differences are easy to program, we have just seen
that this 1D transient diffusion problem is limited to taking rather small
time steps. If, instead, we represent
Eq. :eq:`eq:diffusion:mesh1D:constantD`
as::

    phi_new[j] = phi_old[j] \
      + (D * dt / dx**2) * (phi_new[j+1] - 2 * phi_new[j] + phi_new[j-1])

it is possible to take much larger time steps. Because ``phi_new`` appears on
both the left and right sides of the equation, this form is called
"implicit". In general, the "implicit" representation is much more
difficult to program than the "explicit" form that we just used, but in
:term:`FiPy`, all that is needed is to write

>>> eqI = TransientTerm() == DiffusionTerm(coeff=D)

reset the problem

>>> phi.setValue(valueRight)

and rerun with much larger time steps

>>> timeStepDuration *= 10
>>> steps //= 10
>>> for step in range(steps):
...     eqI.solve(var=phi,
...               dt=timeStepDuration)
...     if __name__ == '__main__':
...         viewer.plot()

>>> print(phi.allclose(phiAnalytical, atol = 2e-2)) # doctest: +SCIPY
1

>>> from fipy import input
>>> if __name__ == '__main__':
...     input("Implicit transient diffusion. Press <return> to proceed...")

.. image:: /figures/examples/diffusion/mesh1Dimplicit.*
   :width: 90%
   :align: center
   :alt: solution to diffusion problem evolved by implicit time steps

Note that although much larger *stable* timesteps can be taken with this
implicit version (there is, in fact, no limit to how large an implicit
timestep you can take for this particular problem), the solution is less
*accurate*. One way to achieve a compromise between *stability* and
*accuracy* is with the Crank-Nicholson scheme, represented by::

    phi_new[j] = phi_old[j] + (D * dt / (2 * dx**2)) * \
                    ((phi_new[j+1] - 2 * phi_new[j] + phi_new[j-1])
                     + (phi_old[j+1] - 2 * phi_old[j] + phi_old[j-1]))

which is essentially an average of the explicit and implicit schemes from
above. This can be rendered in :term:`FiPy` as easily as

>>> eqCN = eqX + eqI

We again reset the problem

>>> phi.setValue(valueRight)

and apply the Crank-Nicholson scheme until the end, when we apply one step
of the fully implicit scheme to drive down the error
(see, *e.g.*, section 19.2 of :cite:`NumericalRecipes`).

>>> for step in range(steps - 1):
...     eqCN.solve(var=phi,
...                dt=timeStepDuration)
...     if __name__ == '__main__':
...         viewer.plot()
>>> eqI.solve(var=phi,
...           dt=timeStepDuration)
>>> if __name__ == '__main__':
...     viewer.plot()

>>> print(phi.allclose(phiAnalytical, atol = 3e-3)) # doctest: +SCIPY
1

>>> from fipy import input
>>> if __name__ == '__main__':
...     input("Crank-Nicholson transient diffusion. Press <return> to proceed...")

----

As mentioned above, there is no stable limit to how large a time step can
be taken for the implicit diffusion problem. In fact, if the time evolution
of the problem is not interesting, it is possible to eliminate the time
step altogether by omitting the :class:`~fipy.terms.transientTerm.TransientTerm`. The steady-state diffusion
equation

.. math::

   D \nabla^2 \phi = 0

is represented in :term:`FiPy` by

>>> DiffusionTerm(coeff=D).solve(var=phi)

>>> if __name__ == '__main__':
...     viewer.plot()

The analytical solution to the steady-state problem is no longer an error
function, but simply a straight line, which we can confirm to a tolerance
of :math:`10^{-10}`.

>>> L = nx * dx
>>> print(phi.allclose(valueLeft + (valueRight - valueLeft) * x / L,
...                    rtol = 1e-10, atol = 1e-10))
1

>>> from fipy import input
>>> if __name__ == '__main__':
...     input("Implicit steady-state diffusion. Press <return> to proceed...")

.. image:: /figures/examples/diffusion/mesh1DsteadyState.*
   :width: 90%
   :align: center
   :alt: steady-state solution to diffusion problem

----

Often, boundary conditions may be functions of another variable in the
system or of time.

For example, to have

.. math::

   \phi = \begin{cases}
       (1 + \sin t) / 2 &\text{on \( x = 0 \)} \\
       0 &\text{on \( x = L \)} \\
   \end{cases}

we will need to declare time :math:`t` as a :class:`~fipy.variables.variable.Variable`

>>> time = Variable()

and then declare our boundary condition as a function of this :class:`~fipy.variables.variable.Variable`

>>> del phi.faceConstraints
>>> valueLeft = 0.5 * (1 + numerix.sin(time))
>>> phi.constrain(valueLeft, mesh.facesLeft)
>>> phi.constrain(0., mesh.facesRight)

>>> eqI = TransientTerm() == DiffusionTerm(coeff=D)

When we update ``time`` at each timestep, the left-hand boundary
condition will automatically update,

>>> dt = .1
>>> while time() < 15:
...     time.setValue(time() + dt)
...     eqI.solve(var=phi, dt=dt)
...     if __name__ == '__main__':
...         viewer.plot()

>>> from fipy import input
>>> if __name__ == '__main__':
...     input("Time-dependent boundary condition. Press <return> to proceed...")

.. image:: /figures/examples/diffusion/mesh1DtimedBC.*
   :width: 90%
   :align: center
   :alt: solution to diffusion problem with a time-dependent Dirichlet boundary condition

----

Many interesting problems do not have simple, uniform diffusivities. We consider a
steady-state diffusion problem

.. math::

   \nabla \cdot ( D \nabla \phi) = 0,

with a spatially varying diffusion coefficient

.. math::

   D = \begin{cases}
   1& \text{for \( 0 < x < L / 4 \),} \\
   0.1& \text{for \( L / 4 \le x < 3 L / 4 \),} \\
   1& \text{for \( 3 L / 4 \le x < L \),}
   \end{cases}

and with boundary conditions
:math:`\phi = 0` at :math:`x = 0` and :math:`D \frac{\partial \phi}{\partial x}
= 1` at :math:`x = L`, where :math:`L` is the length of the solution
domain. Exact numerical answers to this problem are found when the mesh
has cell centers that lie at :math:`L / 4` and :math:`3 L / 4`, or when the
number of cells in the mesh :math:`N_i` satisfies :math:`N_i = 4 i + 2`,
where :math:`i` is an integer. The mesh we've been using thus far is
satisfactory, with :math:`N_i = 50` and :math:`i = 12`.

Because :term:`FiPy` considers diffusion to be a flux from one cell to the next,
through the intervening face, we must define the non-uniform diffusion
coefficient on the mesh faces

.. index::
   single: FaceVariable

>>> D = FaceVariable(mesh=mesh, value=1.0)
>>> X = mesh.faceCenters[0]
>>> D.setValue(0.1, where=(L / 4. <= X) & (X < 3. * L / 4.))

The boundary conditions are a fixed value of

>>> valueLeft = 0.

to the left and a fixed gradient of

>>> gradRight = 1.

to the right:

>>> phi = CellVariable(mesh=mesh, name="solution variable")
>>> phi.faceGrad.constrain([gradRight], mesh.facesRight)
>>> phi.constrain(valueLeft, mesh.facesLeft)

We re-initialize the solution variable

>>> phi.setValue(0)

and obtain the steady-state solution with one implicit solution step

>>> eq = DiffusionTerm(coeff=D)
>>> solver = eq.getDefaultSolver(tolerance=1e-7)
>>> eq.solve(var=phi, solver=solver)

The analytical solution is simply

.. math::

   \phi = \begin{cases}
   x & \text{for \( 0 < x < L/4 \),} \\
   10 x - 9L/4 & \text{for \( L/4 \le x < 3 L / 4 \),} \\
   x + 18 L / 4 & \text{for \( 3 L / 4 \le x < L \),}
   \end{cases}

or

>>> x = mesh.cellCenters[0]
>>> phiAnalytical.setValue(x)
>>> phiAnalytical.setValue(10 * x - 9. * L / 4.,
...                        where=(L / 4. <= x) & (x < 3. * L / 4.))
>>> phiAnalytical.setValue(x + 18. * L / 4.,
...                        where=3. * L / 4. <= x)
>>> print(phi.allclose(phiAnalytical, atol=1e-8, rtol=1e-8))
1

And finally, we can plot the result

>>> from fipy import input
>>> if __name__ == '__main__':
...     Viewer(vars=(phi, phiAnalytical)).plot()
...     input("Non-uniform steady-state diffusion. Press <return> to proceed...")


.. image:: /figures/examples/diffusion/mesh1Dnon-uniform.*
   :width: 90%
   :align: center
   :alt: steady-state solution to diffusion problem with a non-uniform diffusivity

----

Note that for problems involving heat transfer and other similar
conservation equations, it is important to ensure that we begin with
the correct form of the equation. For example, for heat transfer with
:math:`\phi` representing the temperature,

.. math::
    \frac{\partial}{\partial t} \left(\rho \hat{C}_p \phi\right) = \nabla \cdot [ k \nabla \phi ].

With constant and uniform density :math:`\rho`, heat capacity :math:`\hat{C}_p`
and thermal conductivity :math:`k`, this is often written like Eq.
:eq:`eq:diffusion:mesh1D:constantD`, but replacing :math:`D` with :math:`\alpha =
\frac{k}{\rho \hat{C}_p}`. However, when these parameters vary either in position
or time, it is important to be careful with the form of the equation used. For
example, if :math:`k = 1` and

.. math::
   \rho \hat{C}_p = \begin{cases}
   1& \text{for \( 0 < x < L / 4 \),} \\
   10& \text{for \( L / 4 \le x < 3 L / 4 \),} \\
   1& \text{for \( 3 L / 4 \le x < L \),}
   \end{cases},

then we have

.. math::
   \alpha = \begin{cases}
   1& \text{for \( 0 < x < L / 4 \),} \\
   0.1& \text{for \( L / 4 \le x < 3 L / 4 \),} \\
   1& \text{for \( 3 L / 4 \le x < L \),}
   \end{cases}.

However, using a ``DiffusionTerm`` with the same coefficient as that in the
section above is incorrect, as the steady state governing equation reduces to
:math:`0 = \nabla^2\phi`, which results in a linear profile in 1D, unlike that
for the case above with spatially varying diffusivity. Similar care must be
taken if there is time dependence in the parameters in transient problems.

We can illustrate the differences with an example. We define field
variables for the correct and incorrect solution

>>> phiT = CellVariable(name="correct", mesh=mesh)
>>> phiF = CellVariable(name="incorrect", mesh=mesh)
>>> phiT.faceGrad.constrain([gradRight], mesh.facesRight)
>>> phiF.faceGrad.constrain([gradRight], mesh.facesRight)
>>> phiT.constrain(valueLeft, mesh.facesLeft)
>>> phiF.constrain(valueLeft, mesh.facesLeft)
>>> phiT.setValue(0)
>>> phiF.setValue(0)

The relevant parameters are

>>> k = 1.
>>> alpha_false = FaceVariable(mesh=mesh, value=1.0)
>>> X = mesh.faceCenters[0]
>>> alpha_false.setValue(0.1, where=(L / 4. <= X) & (X < 3. * L / 4.))
>>> eqF = 0 == DiffusionTerm(coeff=alpha_false)
>>> eqT = 0 == DiffusionTerm(coeff=k)

>>> eqF.solve(var=phiF)
>>> solver = eqT.getDefaultSolver(tolerance=1e-7)
>>> eqT.solve(var=phiT, solver=solver)

Comparing to the correct analytical solution, :math:`\phi = x`

>>> x = mesh.cellCenters[0]
>>> phiAnalytical.setValue(x)
>>> print(phiT.allclose(phiAnalytical, atol=1e-8, rtol=1e-8)) # doctest: +SCIPY
1

and finally, plot

>>> from fipy import input
>>> if __name__ == '__main__':
...     Viewer(vars=(phiT, phiF)).plot()
...     input("Non-uniform thermal conductivity. Press <return> to proceed...")

.. image:: /figures/examples/diffusion/mesh1Dalpha.*
   :width: 90%
   :align: center
   :alt: representation of difference between non-uniform alpha and D

----

Often, the diffusivity is not only non-uniform, but also depends on
the value of the variable, such that

.. math::
   :label: eq:diffusion:mesh1D:variableD

   \frac{\partial \phi}{\partial t} = \nabla \cdot [ D(\phi) \nabla \phi].

With such a non-linearity, it is generally necessary to "sweep" the
solution to convergence. This means that each time step should be
calculated over and over, using the result of the previous sweep to update
the coefficients of the equation, without advancing in time. In :term:`FiPy`, this
is accomplished by creating a solution variable that explicitly retains its
"old" value by specifying ``hasOld`` when you create it. The variable does
not move forward in time until it is explicitly told to ``updateOld()``. In
order to compare the effects of different numbers of sweeps, let us create
a list of variables: ``phi[0]`` will be the variable that is actually being
solved and ``phi[1]`` through ``phi[4]`` will display the result of taking the
corresponding number of sweeps (``phi[1]`` being equivalent to not sweeping
at all).

>>> valueLeft = 1.
>>> valueRight = 0.
>>> phi = [
...     CellVariable(name="solution variable",
...                  mesh=mesh,
...                  value=valueRight,
...                  hasOld=1),
...     CellVariable(name="1 sweep",
...                  mesh=mesh),
...     CellVariable(name="2 sweeps",
...                  mesh=mesh),
...     CellVariable(name="3 sweeps",
...                  mesh=mesh),
...     CellVariable(name="4 sweeps",
...                  mesh=mesh)
... ]

If, for example,

.. math::

   D = D_0 (1 - \phi)

we would simply write
Eq. :eq:`eq:diffusion:mesh1D:variableD`
as

>>> D0 = 1.
>>> eq = TransientTerm() == DiffusionTerm(coeff=D0 * (1 - phi[0]))

.. note::

   Because of the non-linearity, the Crank-Nicholson scheme does not work
   for this problem.

We apply the same boundary conditions that we used for the uniform
diffusivity cases

>>> phi[0].constrain(valueRight, mesh.facesRight)
>>> phi[0].constrain(valueLeft, mesh.facesLeft)

Although this problem does not have an exact transient solution, it
can be solved in steady-state, with

.. math::

   \phi(x) = 1 - \sqrt{\frac{x}{L}}

>>> x = mesh.cellCenters[0]
>>> phiAnalytical.setValue(1. - numerix.sqrt(x/L))

We create a viewer to compare the different numbers of sweeps with the
analytical solution from before.

>>> if __name__ == '__main__':
...     viewer = Viewer(vars=phi + [phiAnalytical],
...                     datamin=0., datamax=1.)
...     viewer.plot()

As described above, an inner "sweep" loop is generally required for
the solution of non-linear or multiple equation sets. Often a
conditional is required to exit this "sweep" loop given some
convergence criteria. Instead of using the :meth:`~fipy.terms.term.Term.solve`
method equation, when sweeping, it is often useful to call
:meth:`~fipy.terms.term.Term.sweep` instead. The
:meth:`~fipy.terms.term.Term.sweep` method behaves the same way as
:meth:`~fipy.terms.term.Term.solve`, but returns the residual that can then be
used as part of the exit condition.

We now repeatedly run the problem with increasing numbers of
sweeps.

>>> from fipy import input
>>> for sweeps in range(1, 5):
...     phi[0].setValue(valueRight)
...     for step in range(steps):
...         # only move forward in time once per time step
...         phi[0].updateOld()
... 
...         # but "sweep" many times per time step
...         for sweep in range(sweeps):
...             res = eq.sweep(var=phi[0],
...                            dt=timeStepDuration)
...         if __name__ == '__main__':
...             viewer.plot()
... 
...     # copy the final result into the appropriate display variable
...     phi[sweeps].setValue(phi[0])
...     if __name__ == '__main__':
...         viewer.plot()
...         input("Implicit variable diffusivity. %d sweep(s). \
... Residual = %f. Press <return> to proceed..." % (sweeps, (abs(res))))

As can be seen, sweeping does not dramatically change the result, but the
"residual" of the equation (a measure of how accurately it has been solved)
drops about an order of magnitude with each additional sweep.

.. attention::

   Choosing an optimal balance between the number of time steps, the number
   of sweeps, the number of solver iterations, and the solver tolerance is
   more art than science and will require some experimentation on your part
   for each new problem.

Finally, we can increase the number of steps to approach equilibrium, or we
can just solve for it directly

>>> eq = DiffusionTerm(coeff=D0 * (1 - phi[0]))
>>> solver = eq.getDefaultSolver(tolerance=1e-6)

>>> phi[0].setValue(valueRight)
>>> res = 1e+10
>>> while res > 1e-6:
...     res = eq.sweep(var=phi[0],
...                    dt=timeStepDuration,
...                    solver=solver)


>>> print(phi[0].allclose(phiAnalytical, atol = 1e-1))
1

>>> from fipy import input
>>> if __name__ == '__main__':
...     viewer.plot()
...     input("Implicit variable diffusivity - steady-state. \
... Press <return> to proceed...")

.. image:: /figures/examples/diffusion/mesh1Dvariable.*
   :width: 90%
   :align: center
   :alt: solution to a diffusion problem a non-linear diffusivity

----

Fully implicit solutions are not without their pitfalls, particularly in steady
state. Consider a localized block of material diffusing in a closed box.

>>> phi = CellVariable(mesh=mesh, name=r"$\phi$")

>>> phi.value = 0.
>>> phi.setValue(1., where=(x > L/2. - L/10.) & (x < L/2. + L/10.))
>>> if __name__ == '__main__':
...     viewer = Viewer(vars=phi, datamin=-0.1, datamax=1.1)

.. image:: /figures/examples/diffusion/mesh1D-noflux_initial.*
   :width: 90%
   :align: center
   :alt: initial condition for no-flux boundary conditions

We assign no explicit boundary conditions, leaving the default no-flux boundary
conditions, and solve

.. math::

   \partial\phi/\partial t = \nabla\cdot(D\nabla\phi)

>>> D = 1.
>>> eq = TransientTerm() == DiffusionTerm(D)

>>> dt = 10. * dx**2 / (2 * D)
>>> steps = 200

>>> for step in range(steps):
...     eq.solve(var=phi, dt=dt)
...     if __name__ == '__main__':
...         viewer.plot()
>>> from fipy import input
>>> if __name__ == '__main__':
...     input("No-flux - transient. \
... Press <return> to proceed...")

.. image:: /figures/examples/diffusion/mesh1D-noflux_transient.*
   :width: 90%
   :align: center
   :alt: long-time solution for no-flux boundary conditions

and see that :math:`\phi` dissipates to the expected average value of 0.2 with
reasonable accuracy.

>>> print(numerix.allclose(phi, 0.2, atol=4e-5))
True

If we reset the initial condition

>>> phi.value = 0.
>>> phi.setValue(1., where=(x > L/2. - L/10.) & (x < L/2. + L/10.))
>>> if __name__ == '__main__':
...     viewer.plot()

and solve the steady-state problem

>>> try:
...     DiffusionTerm(coeff=D).solve(var=phi)
... except:
...     pass
>>> if __name__ == '__main__':
...     viewer.plot()
>>> from fipy import input
>>> if __name__ == '__main__':
...     input("No-flux - steady-state failure. \
... Press <return> to proceed...")

>>> print(numerix.allclose(phi, 0.2, atol=1e-5)) # doctest: +NOT_TRILINOS_SOLVER
False

.. image:: /figures/examples/diffusion/mesh1D-noflux_steady_fail.*
   :width: 90%
   :align: center
   :alt: (failed) steady-state solution for no-flux boundary conditions

Depending on the solver, we find that the value may be uniformly zero,
infinity, or NaN, or the solver may just fail!
What happened to our no-flux boundary conditions?
Trilinos actually manages to get the correct solution, but this should not
be relied on; this problem has an infinite number of solutions.

The problem is that in the implicit discretization of :math:`\nabla\cdot(D\nabla\phi) = 0`,

.. math::

   \begin{vmatrix}
   \frac{D}{{\Delta x}^2} & -\frac{D}{{\Delta x}^2} & & & & & \\\\[1em]
   \ddots & \ddots & \ddots & & & & \\[1em]
   & -\frac{D}{{\Delta x}^2} & \frac{2 D}{{\Delta x}^2} & -\frac{D}{{\Delta x}^2} & & & \\\\[1em]
   & & -\frac{D}{{\Delta x}^2} & \frac{2 D}{{\Delta x}^2} & -\frac{D}{{\Delta x}^2} & & \\\\[1em]
   & & & -\frac{D}{{\Delta x}^2} & \frac{2 D}{{\Delta x}^2} & -\frac{D}{{\Delta x}^2} & \\\\[1em]
   & & & & \ddots & \ddots & \ddots \\\\[1em]
   & & & & & -\frac{D}{{\Delta x}^2} & \frac{D}{{\Delta x}^2} \\\\[1em]
   \end{vmatrix}
   \begin{vmatrix}
   \phi^\text{new}_{0} \\\\[1em]
   \vdots \\[1em]
   \phi^\text{new}_{j-1} \\\\[1em]
   \phi^\text{new}_{j} \\\\[1em]
   \phi^\text{new}_{j+1} \\\\[1em]
   \vdots \\\\[1em]
   \phi^\text{new}_{N-1}
   \end{vmatrix}
   =
   \begin{vmatrix}
   0 \\\\[1em]
   \vdots \\\\[1em]
   0 \\\\[1em]
   0 \\\\[1em]
   0 \\\\[1em]
   \vdots \\\\[1em]
   0
   \end{vmatrix}

the initial condition :math:`\phi^\text{old}` no longer appears and
:math:`\phi = 0` is a perfectly legitimate solution to this matrix equation.

The solution is to run the transient problem and to take one enormous time step

>>> phi.value = 0.
>>> phi.setValue(1., where=(x > L/2. - L/10.) & (x < L/2. + L/10.))
>>> if __name__ == '__main__':
...     viewer.plot()

>>> eq = (TransientTerm() == DiffusionTerm(D))

The initial residual is much larger than the norm of the right-hand-side
vector, so we use `"initial"` tolerance scaling.

>>> solver = eq.getDefaultSolver(criterion="initial", tolerance=1e-9)
>>> eq.solve(var=phi, dt=1e6*dt, solver=solver)
>>> if __name__ == '__main__':
...     viewer.plot()
>>> from fipy import input
>>> if __name__ == '__main__':
...     input("No-flux - steady-state. \
... Press <return> to proceed...")

>>> print(numerix.allclose(phi, 0.2, atol=1e-5))
True

.. image:: /figures/examples/diffusion/mesh1D-noflux_steady.*
   :width: 90%
   :align: center
   :alt: steady-state solution for no-flux boundary conditions

----

If this example had been written primarily as a script, instead of as
documentation, we would delete every line that does not begin with
either "``>>>``" or "``...``", and then delete those prefixes from the
remaining lines, leaving::

     

     ## This script was derived from
     ## 'examples/diffusion/mesh1D.py'

     nx = 50
     dx = 1.
     mesh = Grid1D(nx = nx, dx = dx)
     phi = CellVariable(name="solution variable",
                        mesh=mesh,
                        value=0)

::

     eq = DiffusionTerm(coeff=D0 * (1 - phi[0]))
     phi[0].setValue(valueRight)
     res = 1e+10
     while res > 1e-6:
         res = eq.sweep(var=phi[0],
                        dt=timeStepDuration)

     print phi[0].allclose(phiAnalytical, atol = 1e-1)
     # Expect:
     # 1
     #
     if __name__ == '__main__':
         viewer.plot()
         input("Implicit variable diffusivity - steady-state. \
     Press <return> to proceed...")

Your own scripts will tend to look like this, although you can always write
them as doctest scripts if you choose.  You can obtain a plain script
like this from some `.../example.py` by typing::

    $ fipy_copy_script .../example.py myExample.py

at the command line.

Most of the :term:`FiPy` examples will be a
mixture of plain scripts and doctest documentation/tests.

.. :term:`FiPy` replace:: `FiPy`

.. .. bibmissing:: /refs.bib
    :sort:
"""

__docformat__ = 'restructuredtext'

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())



##from fipy.tools.profiler.profiler import Profiler
##from fipy.tools.profiler.profiler import calibrate_profiler

"""

This input file again solves a 1D diffusion problem as in
`./examples/diffusion/steadyState/mesh1D/input.py`. The difference
being that the mesh size is given by

    >>> nx = 50
    >>> ny = 50

The result is again tested in the same way:

    >>> eq = DiffusionTerm()
    >>> solver = eq.getDefaultSolver(tolerance=1e-10)
    >>> eq.solve(var, solver=solver)
    >>> Lx = nx * dx
    >>> x = mesh.cellCenters[0]
    >>> analyticalArray = valueLeft + (valueRight - valueLeft) * x / Lx
    >>> print(var.allclose(analyticalArray, atol = 1e-7))
    True

"""

from fipy import input
from fipy import CellVariable, Tri2D, DiffusionTerm, Viewer

nx = 50
ny = 50

dx = 1.

valueLeft = 0.
valueRight = 1.

mesh = Tri2D(dx = dx, nx = nx, ny = ny)

var = CellVariable(name = "solution variable",
                   mesh = mesh,
                   value = valueLeft)

var.constrain(valueLeft, mesh.facesLeft)
var.constrain(valueRight, mesh.facesRight)

if __name__ == '__main__':
    DiffusionTerm().solve(var)
    viewer = Viewer(vars = var)
    viewer.plot()
    input("finished")




r"""Solve the diffusion equation with an anisotropic diffusion coefficient.

We wish to solve the problem

.. math::

   \frac{\partial \phi}{\partial t} = \partial_j \Gamma_{ij} \partial_i \phi

on a circular domain centered at :math:`(0, 0)`. We can choose an anisotropy ratio of 5
such that

.. math::

   \Gamma' = \begin{pmatrix}
       0.2 & 0 \\
         0 & 1
   \end{pmatrix}

A new matrix is formed by rotating :math:`\Gamma'` such that

.. math::

   R = \begin{pmatrix}
        \cos\theta & \sin\theta \\
       -\sin\theta & \cos\theta
   \end{pmatrix}

and

.. math::

   \Gamma = R \Gamma' R^T

In the case of a point source at :math:`(0, 0)` a reference
solution is given by,

.. math::

   \phi \left( X, Y, t \right) = Q \frac{
    \exp \left( -\frac{1}{4 t} \left( \frac{ X^2 }{ \Gamma'_{00}} +
    \frac{ Y^2 }{ \Gamma'_{11}} \right) \right) }{ 4 \pi t
    \sqrt{\Gamma'_{00} \Gamma'_{11}} }

where :math:`\left(X, Y \right)^T = R \left(x, y \right)^T` and :math:`Q` is the initial
mass.

>>> from fipy import CellVariable, Gmsh2D, Viewer, TransientTerm, DiffusionTermCorrection
>>> from fipy.tools import serialComm, numerix

Import a mesh previously created using :term:`Gmsh`.

>>> import os
>>> mesh = Gmsh2D(os.path.splitext(__file__)[0] + '.msh', communicator=serialComm) # doctest: +GMSH

Set the center-most cell to have a value.

>>> var = CellVariable(mesh=mesh, hasOld=1) # doctest: +GMSH
>>> x, y = mesh.cellCenters # doctest: +GMSH
>>> var[numerix.argmin(x**2 + y**2)] = 1. # doctest: +GMSH

Choose an orientation for the anisotropy.

>>> theta = numerix.pi / 4.
>>> rotationMatrix = numerix.array(((numerix.cos(theta), numerix.sin(theta)), \
...                                 (-numerix.sin(theta), numerix.cos(theta))))
>>> gamma_prime = numerix.array(((0.2, 0.), (0., 1.)))
>>> DOT = numerix.NUMERIX.dot
>>> gamma = DOT(DOT(rotationMatrix, gamma_prime), numerix.transpose(rotationMatrix))

Make the equation, viewer and solve.

>>> eqn = TransientTerm() == DiffusionTermCorrection((gamma,))

>>> if __name__ == '__main__':
...     viewer = Viewer(var, datamin=0.0, datamax=0.001)

>>> mass = float(var.cellVolumeAverage * numerix.sum(mesh.cellVolumes)) # doctest: +GMSH
>>> time = 0
>>> dt=0.00025

>>> for i in range(20):
...     var.updateOld() # doctest: +GMSH
...     res = 1.
... 
...     while res > 1e-2:
...         res = eqn.sweep(var, dt=dt) # doctest: +GMSH
... 
...     if __name__ == '__main__':
...         viewer.plot()
...     time += dt

Compare with the analytical solution (within 5% accuracy).

>>> X, Y = numerix.dot(mesh.cellCenters, CellVariable(mesh=mesh, rank=2, value=rotationMatrix)) # doctest: +GMSH
>>> solution = mass * numerix.exp(-(X**2 / gamma_prime[0][0] + Y**2 / gamma_prime[1][1]) / (4 * time)) / (4 * numerix.pi * time * numerix.sqrt(gamma_prime[0][0] * gamma_prime[1][1])) # doctest: +GMSH
>>> print(max(abs((var - solution) / max(solution))) < 0.08) # doctest: +GMSH
True

"""
__docformat__ = 'restructuredtext'

from fipy import input

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())

    input('finished')



r"""Solve a convection problem with a source.

This example solves the equation

.. math::

   \frac{\partial \phi}{\partial x} + \alpha \phi = 0

with :math:`\phi \left( 0 \right) = 1` at :math:`x = 0`. The boundary
condition at :math:`x = L` is an outflow boundary condition requiring
the use of an artificial constraint to be set on the right hand side
faces. Exterior faces without constraints are considered to have zero
outflow. An :class:`~fipy.terms.implicitSourceTerm.ImplicitSourceTerm`
object will be used to represent this term. The derivative of
:math:`\phi` can be represented by a
:class:`~fipy.terms.convectionTerm.ConvectionTerm` with a constant unitary
velocity field from left to right.  The following is an example code that
includes a test against the analytical result.

>>> from fipy import CellVariable, Grid1D, DiffusionTerm, PowerLawConvectionTerm, ImplicitSourceTerm, Viewer
>>> from fipy.tools import numerix

>>> L = 10.
>>> nx = 5000
>>> dx =  L / nx
>>> mesh = Grid1D(dx=dx, nx=nx)
>>> phi0 = 1.0
>>> alpha = 1.0
>>> phi = CellVariable(name=r"$\phi$", mesh=mesh, value=phi0)
>>> solution = CellVariable(name=r"solution", mesh=mesh, value=phi0 * numerix.exp(-alpha * mesh.cellCenters[0]))

>>> from fipy import input
>>> if __name__ == "__main__":
...     viewer = Viewer(vars=(phi, solution))
...     viewer.plot()
...     input("press key to continue")

>>> phi.constrain(phi0, mesh.facesLeft)
>>> ## fake outflow condition
>>> phi.faceGrad.constrain([0], mesh.facesRight)

>>> eq = PowerLawConvectionTerm((1,)) + ImplicitSourceTerm(alpha)
>>> eq.solve(phi)
>>> print(numerix.allclose(phi, phi0 * numerix.exp(-alpha * mesh.cellCenters[0]), atol=1e-3))
True

>>> from fipy import input
>>> if __name__ == "__main__":
...     viewer = Viewer(vars=(phi, solution))
...     viewer.plot()
...     input("finished")


"""
__docformat__ = 'restructuredtext'

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())



"""

This example is a 1D steady state diffusion test case as in
`./examples/diffusion/variable/mesh2x1/input.py` with then
number of cells set to `nx = 10`.

A simple analytical answer can be used to test the result:
   >>> DiffusionTerm(coeff = diffCoeff).solve(var)
   >>> if __name__ == "__main__":
   ...     viewer = Viewer(vars = var)
   ...     viewer.plot()
   >>> x = mesh.cellCenters[0]
   >>> values = numerix.where(x < 3. * L / 4., 10 * x - 9. * L / 4., x + 18. * L / 4.)
   >>> values = numerix.where(x < L / 4., x, values)
   >>> print(var.allclose(values))
   True

"""

from fipy import input
from fipy import FaceVariable, Tri2D, CellVariable, DiffusionTerm, Viewer
from fipy.tools import numerix

nx = 10
ny = 1

valueLeft = 0.
fluxRight = 1.
timeStepDuration = 1.

L = 10.

dx = L / nx
dy = 1.

mesh = Tri2D(dx, dy, nx, ny)

var = CellVariable(
    name = "solution variable",
    mesh = mesh,
    value = valueLeft)

diffCoeff = FaceVariable(mesh = mesh, value = 1.0)

x = mesh.faceCenters[0]
diffCoeff.setValue(0.1, where=(L/4. <= x) & (x < 3. * L / 4.))

var.faceGrad.constrain([[1.], [0.]], mesh.facesRight)

var.constrain(valueLeft, mesh.facesLeft)

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())
    input('finished')




__docformat__ = 'restructuredtext'

from .multilevelPreconditioner import MultilevelPreconditioner

__all__ = ["MultilevelDDPreconditioner"]

class MultilevelDDPreconditioner(MultilevelPreconditioner):
    """Classical smoothed aggregation-based 2-level domain decomposition preconditioner for :class:`~fipy.solvers.trilinos.trilinosSolver.TrilinosSolver`.
    """

    def __init__(self, levels=2):
        """

        Parameters
        ----------
        levels : int
            Maximum number of levels
        """
        self.levels = levels

    @property
    def _parameterList(self):
        return {
            "output": 0,
            "max levels": self.levels,
            "prec type": "MGV",
            "increasing or decreasing": "increasing",
            "aggregation: type": "METIS",
            "aggregation: local aggregates": 1,
            "aggregation: damping factor" : 4. / 3.,
            "eigen-analysis: type": "power-method",
            "eigen-analysis: iterations": 20,
            "smoother: sweeps": 1,
            "smoother: pre or post": "both",
            "smoother: type": "Aztec",
            "smoother: Aztec as solver": False,
            "coarse: type": "Amesos-KLU",
            "coarse: max size": 128
        }



"""
The iterative solvers may output warnings if the solution is considered
unsatisfactory. If you are not interested in these warnings, you can invoke
python with a warning filter such as::

    $ python -Wignore::fipy.SolverConvergenceWarning myscript.py

If you are extremely concerned about your preconditioner for some reason, you
can abort whenever it has problems with::

    $ python -Werror::fipy.PreconditionerWarning myscript.py

"""
__docformat__ = 'restructuredtext'

import logging
import os
import warnings

from fipy.tools import numerix
from .convergence import ConvergenceBase

__all__ = ["SolverConvergenceWarning", "NormalConvergence", "MaximumIterationWarning",
           "PreconditionerWarning", "IllConditionedPreconditionerWarning",
           "PreconditionerNotPositiveDefiniteWarning", "MatrixIllConditionedWarning",
           "StagnatedSolverWarning", "ScalarQuantityOutOfRangeWarning",
           "IllegalInputOrBreakdownWarning",
           "ParameterWarning", "BreakdownWarning", "LossOfPrecisionWarning",
           "Solver"]

class SolverConvergenceWarning(Warning):
    def __init__(self, solver, iter, relres):
        self.solver = solver
        self.iter = iter
        self.relres = relres

    def __str__(self):
        return "%s failed. Iterations: %g. Relative error: %g" % (str(self.solver), self.iter, self.relres)

class NormalConvergence(SolverConvergenceWarning):
    def __str__(self):
        return "User requested convergence criteria is satisfied. Iterations: {0}. Relative error: {1}".format(self.iter, self.relres)

class MaximumIterationWarning(SolverConvergenceWarning):
    def __str__(self):
        return "Iterations: %g. Relative error: %g" % (self.iter, self.relres)

class PreconditionerWarning(SolverConvergenceWarning):
    pass

class IllConditionedPreconditionerWarning(PreconditionerWarning):
    def __str__(self):
        return "The system involving the preconditioner was ill-conditioned. Relative error: %g" % (self.relres)

class PreconditionerNotPositiveDefiniteWarning(PreconditionerWarning):
    def __str__(self):
        return "The preconditioning matrix does not appear to be positive definite. Relative error: %g" % (self.relres)

class MatrixIllConditionedWarning(SolverConvergenceWarning):
    def __str__(self):
        return "The matrix appears to be very ill-conditioned. Relative error: %g" % (self.relres)

class StagnatedSolverWarning(SolverConvergenceWarning):
    def __str__(self):
        return "The solver stagnated. Iterations: %g. Relative error: %g" % (self.iter, self.relres)

class ScalarQuantityOutOfRangeWarning(SolverConvergenceWarning):
    def __str__(self):
        return "A scalar quantity became too small or too large to continue computing. Iterations: %g. Relative error: %g" % (self.iter, self.relres)

class IllegalInputOrBreakdownWarning(SolverConvergenceWarning):
    def __str__(self):
        return "{0} received illegal input or had a breakdown." \
          "Iterations: {1}. Relative error: {2}".format(self.solver, self.iter, self.relres)

class ParameterWarning(SolverConvergenceWarning):
    def __str__(self):
        return "User requested option is not available for {0}.".format(self.solver)

class BreakdownWarning(SolverConvergenceWarning):
    def __str__(self):
        return "Numerical breakdown occurred. Iterations: {0}. Relative error: {1}".format(self.iter, self.relres)

class LossOfPrecisionWarning(SolverConvergenceWarning):
    def __str__(self):
        return "Numerical loss of precision occurred. Iterations: {0}. Relative error: {1}".format(self.iter, self.relres)



class Solver(object):
    """
    The base `LinearXSolver` class.

    .. attention:: This class is abstract. Always create one of its subclasses.
    """

    #: Default tolerance for linear solves unless `criterion="legacy"`
    DEFAULT_TOLERANCE = 1e-5

    #: Default tolerance for linear solves if `criterion="legacy"`
    LEGACY_TOLERANCE = 1e-10

    #: Default maximum number of iterative steps to perform
    DEFAULT_ITERATIONS = 1000

    #: Default preconditioner to apply to the matrix
    DEFAULT_PRECONDITIONER = None

    def __init__(self, tolerance="default", criterion="default",
                 iterations="default", precon="default"):
        """
        Create a `Solver` object.

        Parameters
        ----------
        tolerance : float
            Required residual tolerance.
        criterion : {'default', 'initial', 'unscaled', 'RHS', 'matrix', 'solution', 'preconditioned', 'natural', 'legacy'}, optional
            Interpretation of ``tolerance``.
            See :ref:`CONVERGENCE` for more information.
        iterations : int
            Maximum number of iterative steps to perform.
        precon : ~fipy.solvers.preconditioner.Preconditioner
            Preconditioner to use.  Not all solver suites support
            preconditioners.
        """
        if self.__class__ is Solver:
            raise NotImplementedError("can't instantiate abstract base class")

        self.criterion = self.value_or_default(criterion,
                                               "default",
                                               "FIPY_DEFAULT_CRITERION")
        self.tolerance = self.value_or_default(tolerance,
                                               self.default_tolerance)
        self.iterations = self.value_or_default(iterations,
                                                self.DEFAULT_ITERATIONS)
        self.preconditioner = self.value_or_default(precon,
                                                    self.default_preconditioner)

        self._log = logging.getLogger(self.__class__.__module__
                                      + "." + self.__class__.__name__)

    def value_or_default(self, value, default, envvar=None):
        if value == "default":
            if envvar is not None:
                value = os.environ.get(envvar, default)
            else:
                value = default
        return value

    @property
    def default_tolerance(self):
        """Default tolerance for linear solve
        """
        if self.criterion == "legacy":
            return self.LEGACY_TOLERANCE
        else:
            return self.DEFAULT_TOLERANCE

    @property
    def default_preconditioner(self):
        if self.DEFAULT_PRECONDITIONER is not None:
            # instantiate DEFAULT_PRECONDITIONER class
            return self.DEFAULT_PRECONDITIONER()
        else:
            return None

    def _storeMatrix(self, var, matrix, RHSvector):
        self.var = var
        self.matrix = matrix
        self.RHSvector = RHSvector

    def _scatterGhosts(self, x):
        """Distribute ghost values (if any) across processes
        """
        return x

    def _cleanup(self):
        pass

    def _solve(self):
        """Solve system of equations posed for FiPy

        Common method invoked by :class:`~fipy.terms.term.Term`, which then
        calls solver-suite-specific :math:`~fipy.solvers.solver._solve_`
        methods.
        """
        L, x, b = self._Lxb

        if self._log.isEnabledFor(logging.DEBUG):
            s = "Lnorm: {Lnorm}, bnorm: {bnorm}, rnorm: {rnorm}"
            self._log.debug(s.format(Lnorm=self._matrixNorm(L, x, b),
                                     bnorm=self._rhsNorm(L, x, b),
                                     rnorm=self._residualNorm(L, x, b)))

        x = self._solve_(L, x, b)

        x = self._scatterGhosts(x)

        factor = self.var.unit.factor
        if factor != 1:
            x /= self.var.unit.factor

        self.var.value = x.reshape(self.var.shape)

        self._cleanup()

    def _solve_(self, L, x, b):
        """Solve system of equations posed for solver suite

        Parameters
        ----------
        L : ~fipy.matrices.sparseMatrix._SparseMatrix
            Sparse matrix object
        x : array_like
            Solution variable in form suitable for solver
        b : array_like
            Right-hand side vector in form suitable for solver

        Returns
        -------
        ndarray
            Solution vector
        """
        raise NotImplementedError

    def _setConvergence(self, suite, code, iterations, residual, actual_code=None, **kwargs):
        cls = ConvergenceBase.code_registry[(suite, code)]
        self.convergence = cls(solver=self,
                               iterations=iterations,
                               residual=residual,
                               criterion=self.criterion,
                               actual_code=actual_code,
                               **kwargs)

    def _legacyNorm(self, L, x, b):
        raise NotImplementedError

    def _unscaledNorm(self, L, x, b):
        return 1.

    def _rhsNorm(self, L, x, b):
        raise NotImplementedError

    def _matrixNorm(self, L, x, b):
        raise NotImplementedError

    def _residualVectorAndNorm(self, L, x, b):
        raise NotImplementedError

    def _residualNorm(self, L, x, b):
        _, residual = self._residualVectorAndNorm(L, x, b)

        return residual

    def _solutionNorm(self, L, x, b):
        raise NotImplementedError

    def _preconditionedNorm(self, L, x, b):
        raise NotImplementedError

    def _naturalNorm(self, L, x, b):
        raise NotImplementedError

    @property
    def _Lxb(self):
        """Matrix, solution vector, and right-hand side vector

        Returns
        -------
        L : matrix
            Sparse matrix in form suitable for solver
        x : ndarray
            Solution variable in form suitable for solver
        b : ndarray
            Right-hand side vector in form suitable for solver
        """
        if self.var.mesh.communicator.Nproc > 1:
            raise Exception(str(type(self)) + " cannot be used with multiple processors")

        L = self.matrix.matrix
        x = self.var.numericValue.ravel()
        b = numerix.asarray(self.RHSvector)

        if ((self.matrix == 0)
            or (L.shape[0] != L.shape[1])
            or (L.shape[0] != len(x))):

            from fipy.terms import SolutionVariableNumberError

            raise SolutionVariableNumberError

        return (L, x, b)

    def _adaptLegacyTolerance(self, L, x, b):
        raise NotImplementedError

    def _adaptUnscaledTolerance(self, L, x, b):
        raise NotImplementedError

    def _adaptRHSTolerance(self, L, x, b):
        raise NotImplementedError

    def _adaptMatrixTolerance(self, L, x, b):
        raise NotImplementedError

    def _adaptInitialTolerance(self, L, x, b):
        raise NotImplementedError

    def _adaptSolutionTolerance(self, L, x, b):
        raise NotImplementedError

    def _adaptPreconditionedTolerance(self, L, x, b):
        raise NotImplementedError

    def _adaptNaturalTolerance(self, L, x, b):
        raise NotImplementedError

    def _adaptTolerance(self, L, x, b):
        adapt = {
            "legacy": self._adaptLegacyTolerance,
            "unscaled": self._adaptUnscaledTolerance,
            "RHS": self._adaptRHSTolerance,
            "matrix": self._adaptMatrixTolerance,
            "initial": self._adaptInitialTolerance,
            "solution": self._adaptSolutionTolerance,
            "preconditioned": self._adaptPreconditionedTolerance,
            "natural": self._adaptNaturalTolerance,
            "default": self._adaptRHSTolerance
        }

        tolerance_scale, suite_criterion = adapt[self.criterion](L, x, b)

        return tolerance_scale, suite_criterion

    @staticmethod
    def scale_tolerance(tol, scale):
        if tol is not None:
            tol *= scale
        return tol

    def _applyUnderRelaxation(self, underRelaxation=None):
        if underRelaxation is not None:
            self.matrix.putDiagonal(numerix.asarray(self.matrix.takeDiagonal()) / underRelaxation)
            self.RHSvector += (1 - underRelaxation) * self.matrix.takeDiagonal() * numerix.array(self.var).flatten()

    def _calcResidualVector(self, residualFn=None):
        if residualFn is not None:
            return residualFn(self.var, self.matrix, self.RHSvector)
        else:
            Lx = self.matrix * numerix.array(self.var).flatten()

            return Lx - self.RHSvector

    def _calcResidual(self, residualFn=None):
        if residualFn is not None:
            return residualFn(self.var, self.matrix, self.RHSvector)
        else:
            return numerix.L2norm(self._calcResidualVector())

    def _calcRHSNorm(self):
        return numerix.L2norm(self.RHSvector)

    def __repr__(self):
        return '%s(tolerance=%g, iterations=%g)' \
            % (self.__class__.__name__, self.tolerance, self.iterations)

    def _canSolveAsymmetric(self):
        return True

    def __enter__(self):
        # The __enter__() and __exit__() methods allow
        # solver objects to be used within a context manager.
        # This enables one to write:
        #
        #     with ExampleSolver() as solver:
        #         eq.solve(var=phi, solver=solver)
        #
        # For subclasses that do not define __enter__()
        # and __exit__() methods, the above is equivalent to:
        #
        #     eq.solve(var=phi, solver=ExampleSolver())

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass


    def __del__(self):
        pass

    def _test(self):
        r"""
        >>> import fipy as fp

        For sufficiently constrained circumstances, all solver suites
        should do the same thing.  The following problem setup is designed
        to ensure that all interpret solver criteria correctly and achieve
        the "same" tolerance in the same number of iterations.

        Consider a steady-state 1D diffusion problem with a
        position-dependent diffusivity and Dirichlet boundary conditions:

        .. math::

           \begin{aligned}
           \frac{\partial}{\partial x}\left[
               \left(1 + x\right)
               \frac{\partial \phi}{\partial x}
           \right] &= 0
           \\
           \left.\phi\right\rvert_{x=0} &= \phi_L
           \\
           \left.\phi\right\rvert_{x=1} &= \phi_R
           \end{aligned}

        with the analytical solution

        .. math::

           \phi = \frac{\phi_R - \phi_L}{\ln 2} \ln\left(1 + x\right) + \phi_L

        >>> N = 100
        >>> mesh = fp.Grid1D(nx=N, Lx=1)
        >>> phi = fp.CellVariable(mesh=mesh, name=r"$\phi")
        >>> phiL = 1000.
        >>> phiR = 2000.
        >>> phi_analytical = ((((phiR - phiL)/fp.numerix.log(2.))
        ...                    * fp.numerix.log(1 + mesh.x))
        ...                   + phiL)
        >>> phi_analytical.name = r"$\phi_\mathrm{analytical}$"

        >>> fp.numerix.random.seed(12345)
        >>> variance = 1e-3
        >>> phi_initial = phi_analytical + fp.GaussianNoiseVariable(mesh=mesh, variance=variance)
        >>> phi.value = phi_initial
        >>> phi.constrain(phiL, where=mesh.facesLeft)
        >>> phi.constrain(phiR, where=mesh.facesRight)
        >>> D = fp.FaceVariable(mesh=mesh, value=1 + mesh.faceCenters[0])
        >>> eq = fp.DiffusionTerm(coeff=D) == 0

        For reproducibility between suites, we select a solver with
        predictable characteristics (that counts out GMRES) and no
        preconditioning.

        >>> Solver = fp.LinearCGSSolver
        >>> solver = Solver(precon=None)

        >>> solver = eq._prepareLinearSystem(var=phi,
        ...                                  solver=solver,
        ...                                  boundaryConditions=(),
        ...                                  dt=1.)
        >>> L, x, b = solver._Lxb

        The problem parameters were chosen to give good separation between the
        different convergence norms.

        The norm of the matrix is the infinity norm

        .. math::

           \left\| L_{ij}\right\|_\infty &= \max_i \sum_j \left| A_ij \right|
           \\
           &= \max_i \left[
               \left| -N(1 + x_i) \right|
               + \left| 2N(1 + x_i) \right|
               + \left| -N(1 + x_i) \right|
           \right]
           \\
           &= \max_i 4N(1 + x_i)
           &= \mathcal{O}(8 N)

        >>> Lnorm = solver._matrixNorm(L, x, b)
        >>> print(numerix.allclose(Lnorm, 8 * N, rtol=0.1))
        True

        The right-hand-side vector is zero except at the boundaries,
        where the contribution is

        .. math::

           \frac{(1 + x) \phi_{BC} A_f}{d_{AP}} &= (1 + x) \phi_{BC} 2 N
           \\
           &= 2 N \phi_L = 2000 N\qquad\text{at $x = 0$}
           \\
           &= 4 N \phi_R = 8000 N\qquad\text{at $x = 1$}

        Thus the :math:`L_2` norm of the right-hand-side vector is
        :math:`\left\| b \right\|_2 = \mathcal{O}(8000 N)`.

        >>> bnorm = solver._rhsNorm(L, x, b)
        >>> print(numerix.allclose(bnorm, 8000 * N, rtol=0.1))
        True

        We choose the initial condition such that the initial residual will
        be small.

        .. math::

           \phi_0 &= \phi_\text{analytical} + \mathcal{O}(\sigma)
           \\
           r = L \phi_0 - b
           &= L \phi_\text{analytical} - b + L \mathcal{O}(\sigma)
           \\
           &= L \mathcal{O}(\sigma)
           \\
           \left\| r \right\|_2 &= \left\| L \mathcal{O}(\sigma) \right\|_2
           \\
           &= \sqrt{\sum_{0 \le i < N} \left[
               N(1 + x_i) \mathcal{O}(\sigma)
               + 2N(1 + x_i) \mathcal{O}(\sigma)
               + N(1 + x_i) \mathcal{O}(\sigma)
           \right]^2}
           \\
           &= 4 N \mathcal{O}(\sigma) \sqrt{\sum_{0 \le i < N} (1 + x_i)^2}
           \\
           &= \text{probably $\sqrt{\pi}$ or something}
           \\
           &= \mathcal{O}(4 N \sqrt{N} \sigma)

        >>> rnorm = solver._residualNorm(L, x, b)
        >>> print(numerix.allclose(rnorm, 4 * N * numerix.sqrt(N * variance),
        ...                        rtol=0.1))
        True

        Calculate the error of the initial condition (probably could be
        estimated via truncation error blah blah blah).

        >>> enorm = fp.numerix.L2norm(phi - phi_analytical) / fp.numerix.L2norm(phi_analytical)

        >>> from fipy.solvers.convergence import Convergence

        Check that:
        - the solution is converged,
        - the solver reaches the desired residual for the
          criterion, without overshooting too much.  Most get close, but
          "unscaled" overshoots a lot for most suites.
        - the iteration count is as expected
        - the error has been reduced from the initial guess

        PETSc and Trilinos, in parallel, can both take an extra iteration for unscaled

        >>> criteria = {
        ...     "unscaled": (1., 0.003, 120, 114),
        ...     "RHS": (bnorm, 0.6, 2, 2),
        ...     "matrix": (Lnorm, 0.6, 63, 58),
        ...     "initial": (rnorm, 0.6, 114, 105)
        ... }

        >>> # criteria["solution"] = ???  doctest: +TRILINOS_SOLVER
        >>> criteria["preconditioned"] = (bnorm, 0.6, 2, 2) # doctest: +PETSC_SOLVER
        >>> criteria["natural"] = (bnorm, 0.6, 6, 6) # doctest: +PETSC_SOLVER

        >>> satisfied = {}
        >>> for criterion, (target, lower_bound, iter_upper, iter_lower) in criteria.items():
        ...     phi.setValue(phi_initial)
        ...     with Solver(criterion=criterion, precon=None) as s:
        ...         res = eq.sweep(var=phi, solver=s)
        ...         error = (fp.numerix.L2norm(phi - phi_analytical)
        ...                  / fp.numerix.L2norm(phi_analytical))
        ...         checks = [isinstance(s.convergence, Convergence),
        ...                   (lower_bound
        ...                    < (s.convergence.residual
        ...                       / (s.tolerance * target))
        ...                    < 1.0),
        ...                   (iter_lower
        ...                    <= s.convergence.iterations
        ...                    <= iter_upper),
        ...                   error < enorm]
        ...         satisfied[criterion] = all(checks)
        >>> print(all(satisfied.values()))
        True
        """
        pass

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()



r"""
In this example, we present the same three-component diffusion problem
introduced in ``examples/elphf/diffusion/mesh1D.py``
but we demonstrate FiPy's facility to use dimensional quantities.

    >>> import warnings
    >>> warnings.warn("\n\n\tSupport for physical dimensions is incomplete.\n\tIt is not possible to solve dimensional equations.\n")

    >>> from fipy import CellVariable, FaceVariable, PhysicalField, Grid1D, TransientTerm, DiffusionTerm, PowerLawConvectionTerm, LinearLUSolver, Viewer
    >>> from fipy.tools import numerix

We solve the problem on a 40 mm long 1D mesh

    >>> nx = 40
    >>> dx = PhysicalField(1., "mm")
    >>> L = nx * dx
    >>> mesh = Grid1D(dx = dx, nx = nx)

Again, one component in this ternary system will be designated the "solvent"

    >>> class ComponentVariable(CellVariable):
    ...     def __init__(self, mesh, value = 0., name = '',
    ...                  standardPotential = 0., barrier = 0.,
    ...                  diffusivity = None, valence = 0, equation = None):
    ...         CellVariable.__init__(self, mesh = mesh, value = value,
    ...                               name = name)
    ...         self.standardPotential = Variable(standardPotential)
    ...         self.barrier = Variable(barrier)
    ...         self.diffusivity = Variable(diffusivity)
    ...         self.valence = valence
    ...         self.equation = equation
    ... 
    ...     def copy(self):
    ...         return self.__class__(mesh = self.mesh,
    ...                               value = self.value,
    ...                               name = self.name,
    ...                               standardPotential =
    ...                                   self.standardPotential,
    ...                               barrier = self.barrier,
    ...                               diffusivity = self.diffusivity,
    ...                               valence = self.valence,
    ...                               equation = self.equation)

    >>> solvent = ComponentVariable(mesh = mesh, name = 'Cn', value = "1 mol/m**3")

We can create an arbitrary number of components,
simply by providing a `Tuple` or `list` of components

    >>> substitutionals = [
    ...     ComponentVariable(mesh = mesh, name = 'C1', diffusivity = "1e-9 m**2/s",
    ...                       standardPotential = 1., barrier = 1., value = "0.3 mol/m**3"),
    ...     ComponentVariable(mesh = mesh, name = 'C2', diffusivity = "1e-9 m**2/s",
    ...                       standardPotential = 1., barrier = 1., value = "0.6 mol/m**3"),
    ...     ]

    >>> interstitials = []

    >>> for component in substitutionals:
    ...     solvent -= component

We separate the solution domain into two different concentration regimes

    >>> x = mesh.cellCenters[0]
    >>> substitutionals[0].setValue("0.3 mol/m**3")
    >>> substitutionals[0].setValue("0.6 mol/m**3", where=x > L / 2)
    >>> substitutionals[1].setValue("0.6 mol/m**3")
    >>> substitutionals[1].setValue("0.3 mol/m**3", where=x > L / 2)

We create one diffusion equation for each substitutional component

    >>> for Cj in substitutionals:
    ...     CkSum = ComponentVariable(mesh = mesh, value = 0.)
    ...     CkFaceSum = FaceVariable(mesh = mesh, value = 0.)
    ...     for Ck in [Ck for Ck in substitutionals if Ck is not Cj]:
    ...         CkSum += Ck
    ...         CkFaceSum += Ck.harmonicFaceValue
    ... 
    ...     convectionCoeff = CkSum.faceGrad \
    ...                       * (Cj.diffusivity / (1. - CkFaceSum))
    ... 
    ...     Cj.equation = (TransientTerm()
    ...                    == DiffusionTerm(coeff=Cj.diffusivity)
    ...                    + PowerLawConvectionTerm(coeff = convectionCoeff))

If we are running interactively, we create a viewer to see the results

    >>> if __name__ == '__main__':
    ...     viewer = Viewer(vars=[solvent] + substitutionals,
    ...                     datamin=0, datamax=1)
    ...     viewer.plot()

Now, we iterate the problem to equilibrium, plotting as we go

    >>> solver = LinearLUSolver()

    >>> for i in range(40):
    ...     for Cj in substitutionals:
    ...         Cj.updateOld()
    ...     for Cj in substitutionals:
    ...         Cj.equation.solve(var = Cj,
    ...                           dt = "1000 s",
    ...                           solver = solver)
    ...     if __name__ == '__main__':
    ...         viewer.plot()

Since there is nothing to maintain the concentration separation in this problem,
we verify that the concentrations have become uniform

    >>> print(substitutionals[0].scaled.allclose("0.45 mol/m**3",
    ...     atol = "1e-7 mol/m**3", rtol = 1e-7))
    1
    >>> print(substitutionals[1].scaled.allclose("0.45 mol/m**3",
    ...     atol = "1e-7 mol/m**3", rtol = 1e-7))
    1

.. note::

   The absolute tolerance `atol` must be in units compatible with the value to
   be checked, but the relative tolerance `rtol` is dimensionless.
"""
__docformat__ = 'restructuredtext'

from fipy import input

if __name__ == '__main__':
    ## from fipy.tools.profiler.profiler import Profiler
    ## from fipy.tools.profiler.profiler import calibrate_profiler

    # fudge = calibrate_profiler(10000)
    # profile = Profiler('profile', fudge=fudge)

    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())

    # profile.stop()

    input("finished")




from scipy.sparse import csr_matrix, linalg

import pyamgx

from fipy.solvers.solver import Solver
from fipy.matrices.scipyMatrix import _ScipyMeshMatrix
from fipy.tools import numerix
from fipy.tools.timer import Timer

__all__ = ["PyAMGXSolver"]

class PyAMGXSolver(Solver):

    # AMGX configuration options
    CONFIG_DICT = {}

    #: Default smoother to apply to the ???
    DEFAULT_SMOOTHER = None

    def __init__(self, tolerance="default", criterion="default",
                 iterations="default", precon="default", smoother="default", **kwargs):
        """
        Parameters
        ----------
        tolerance : float
            Required error tolerance.
        criterion : {'default', 'unscaled', 'RHS', 'matrix', 'initial', 'legacy'}
            Interpretation of ``tolerance``.
            See :ref:`CONVERGENCE` for more information.
        iterations : int
            Maximum number of iterative steps to perform.
        precon : ~fipy.solvers.pyamgx.preconditioners.PyAMGXPreconditioner, optional
        smoother : ~fipy.solvers.pyamgx.smoothers.Smoother, optional
        **kwargs
            Other AMGX solver options
        """
        super(PyAMGXSolver, self).__init__(tolerance=tolerance, criterion=criterion, iterations=iterations)

        # update solver config:
        self.config_dict = self.CONFIG_DICT.copy()

        self.config_dict["solver"]["max_iters"] = self.iterations

        if self.precon is not None:
            self.precon._applyToSolver(self.config_dict["solver"])

        smoother = self.value_or_default(smoother, self.default_smoother)
        if smoother is not None:
            smoother._applyToSolver(self.config_dict["solver"])

        self.config_dict["solver"].update(kwargs)

        # create AMGX objects:
        self.cfg = pyamgx.Config().create_from_dict(self.config_dict)
        self.resources = pyamgx.Resources().create_simple(self.cfg)
        self.x_gpu = pyamgx.Vector().create(self.resources)
        self.b_gpu = pyamgx.Vector().create(self.resources)
        self.A_gpu = pyamgx.Matrix().create(self.resources)

    @property
    def default_smoother(self):
        if self.DEFAULT_SMOOTHER is not None:
            # instantiate DEFAULT_SMOOTHER class
            return self.DEFAULT_SMOOTHER()
        else:
            return None

    def _destroy_AMGX(self):
        # destroy AMGX objects:
        # self.resources apparently doesn't need to be destroyed
        if hasattr(self, "A_gpu"):
            self.A_gpu.destroy()
            del self.A_gpu
        if hasattr(self, "b_gpu"):
            self.b_gpu.destroy()
            del self.b_gpu
        if hasattr(self, "x_gpu"):
            self.x_gpu.destroy()
            del self.x_gpu
        if hasattr(self, "cfg"):
            self.cfg.destroy()
            del self.cfg

    def __exit__(self, *args):
        self._destroy_AMGX()

    def __del__(self):
        self._destroy_AMGX()

    @property
    def _matrixClass(self):
        return _ScipyMeshMatrix

    def _rhsNorm(self, L, x, b):
        return numerix.L2norm(b)

    def _matrixNorm(self, L, x, b):
        return linalg.norm(L.matrix, ord=numerix.inf)

    def _adaptLegacyTolerance(self, L, x, b):
        return self._adaptInitialTolerance(L, x, b)

    def _adaptUnscaledTolerance(self, L, x, b):
        return (1., "ABSOLUTE")

    def _adaptRHSTolerance(self, L, x, b):
        return (self._rhsNorm(L, x, b), "ABSOLUTE")

    def _adaptMatrixTolerance(self, L, x, b):
        return (self._matrixNorm(L, x, b), "ABSOLUTE")

    def _adaptInitialTolerance(self, L, x, b):
        return (1., "RELATIVE_INI_CORE")

    def _solve_(self, L, x, b):
        # transfer data from CPU to GPU

        self._log.debug("BEGIN cpu2gpu")

        with Timer() as t:
            self.x_gpu.upload(x)
            self.b_gpu.upload(b)
            self.A_gpu.upload_CSR(L)

        self._log.debug("END cpu2gpu - {elapsed} ns".format(elapsed=t.elapsed))

        tolerance_scale, suite_criterion = self._adaptTolerance(L, x, b)
        config_dict = self.config_dict.copy()
        config_dict["solver"]["monitor_residual"] = 1
        config_dict["solver"]["store_res_history"] = 1
        config_dict["solver"]["tolerance"] = self.tolerance * tolerance_scale
        config_dict["solver"]["convergence"] = suite_criterion

        cfg = pyamgx.Config().create_from_dict(config_dict)
        solver = pyamgx.Solver().create(self.resources, cfg)
        solver.setup(self.A_gpu)

        # solve system on GPU

        self._log.debug("BEGIN solve")

        with Timer() as t:
            solver.solve(self.b_gpu, self.x_gpu)

        self._log.debug("END solve - {} ns".format(t.elapsed))

        # download values from GPU to CPU

        self._log.debug("BEGIN gpu2cpu")

        with Timer() as t:
            self.x_gpu.download(x)

        self._log.debug("END gpu2cpu - {} ns".format(t.elapsed))

        if solver.iterations_number == -1:
            residual = None
        else:
            residual = solver.get_residual()

        self._setConvergence(suite="pyamgx",
                             code=solver.status,
                             iterations=solver.iterations_number,
                             tolerance_scale=tolerance_scale,
                             residual=residual)

        self.convergence.warn()

        solver.destroy()
        cfg.destroy()

        return x



##from fipy.tools.profiler.profiler import Profiler
##from fipy.tools.profiler.profiler import calibrate_profiler

"""

This input file again solves a 1D diffusion problem as in
`./examples/diffusion/steadyState/mesh1D/input.py`. In order to test
the non-orthogonality error, this uses a `SkewedGrid2D`, which is a
`Grid2D` with each interior vertex moved in a random direction.

"""

from fipy import input
if __name__ == '__main__':
    import sys
    import os

    from fipy import SkewedGrid2D, CellVariable, Viewer, DiffusionTerm
    from fipy.tools import numerix

    valueLeft = 0.
    valueRight = 1.

    mesh = SkewedGrid2D(dx = 1.0, dy = 1.0, nx = 20, ny = 20, rand = 0.1)

    var = CellVariable(name = "solution variable",
                       mesh = mesh,
                       value = valueLeft)

    viewer = Viewer(vars = var)

    var.constrain(valueLeft, mesh.facesLeft)
    var.constrain(valueRight, mesh.facesRight)

    DiffusionTerm().solve(var)

    varArray = numerix.array(var)
    x = mesh.cellCenters[0]
    analyticalArray = valueLeft + (valueRight - valueLeft) * x / 20
    errorArray = varArray - analyticalArray
    errorVar = CellVariable(name = 'absolute error',
                            mesh = mesh,
                            value = abs(errorArray))
    errorViewer = Viewer(vars = errorVar)

    NonOrthoVar = CellVariable(name = "non-orthogonality",
                               mesh = mesh,
                               value = mesh._nonOrthogonality)
    NOViewer = Viewer(vars = NonOrthoVar)
    viewer.plot()
    NOViewer.plot()

    input("finished")



"""

Input file for chemotaxis modeling.

Here are some test cases for the model.


>>> from examples.chemotaxis.parameters import parameters
>>> from fipy import CellVariable, Grid1D, TransientTerm, DiffusionTerm, ImplicitSourceTerm, Viewer, numerix

>>> params = parameters['case 2']

>>> nx = 50
>>> dx = 1.
>>> L = nx * dx

>>> mesh = Grid1D(nx=nx, dx=dx)

>>> shift = 1.

>>> KMVar = CellVariable(mesh=mesh, value=params['KM'] * shift, hasOld=1)
>>> KCVar = CellVariable(mesh=mesh, value=params['KC'] * shift, hasOld=1)
>>> TMVar = CellVariable(mesh=mesh, value=params['TM'] * shift, hasOld=1)
>>> TCVar = CellVariable(mesh=mesh, value=params['TC'] * shift, hasOld=1)
>>> P3Var = CellVariable(mesh=mesh, value=params['P3'] * shift, hasOld=1)
>>> P2Var = CellVariable(mesh=mesh, value=params['P2'] * shift, hasOld=1)
>>> RVar = CellVariable(mesh=mesh, value=params['R'], hasOld=1)

>>> PN = P3Var + P2Var

>>> KMscCoeff = params['chiK'] * (RVar + 1) * (1 - KCVar - KMVar.cellVolumeAverage)
>>> KMspCoeff = params['lambdaK'] / (1 + PN / params['kappaK'])
>>> KMEq = TransientTerm() - KMscCoeff + ImplicitSourceTerm(KMspCoeff)

>>> TMscCoeff = params['chiT'] * (1 - TCVar - TMVar.cellVolumeAverage)
>>> TMspCoeff = params['lambdaT'] * (KMVar + params['zetaT'])
>>> TMEq = TransientTerm() - TMscCoeff + ImplicitSourceTerm(TMspCoeff)

>>> TCscCoeff = params['lambdaT'] * (TMVar * KMVar).cellVolumeAverage
>>> TCspCoeff = params['lambdaTstar']
>>> TCEq = TransientTerm() - TCscCoeff + ImplicitSourceTerm(TCspCoeff)

>>> PIP2PITP = PN / (PN / params['kappam'] + PN.cellVolumeAverage / params['kappac'] + 1) + params['zetaPITP']

>>> P3spCoeff = params['lambda3'] * (TMVar + params['zeta3T'])
>>> P3scCoeff = params['chi3'] * KMVar * (PIP2PITP / (1 + KMVar / params['kappa3']) + params['zeta3PITP']) + params['zeta3']
>>> P3Eq = TransientTerm() - DiffusionTerm(params['diffusionCoeff']) - P3scCoeff + ImplicitSourceTerm(P3spCoeff)

>>> P2scCoeff = scCoeff = params['chi2'] + params['lambda3'] * params['zeta3T'] * P3Var
>>> P2spCoeff = params['lambda2'] * (TMVar + params['zeta2T'])
>>> P2Eq = TransientTerm() - DiffusionTerm(params['diffusionCoeff']) - P2scCoeff + ImplicitSourceTerm(P2spCoeff)

>>> KCscCoeff = params['alphaKstar'] * params['lambdaK'] * (KMVar / (1 + PN / params['kappaK'])).cellVolumeAverage
>>> KCspCoeff = params['lambdaKstar'] / (params['kappaKstar'] + KCVar)
>>> KCEq = TransientTerm() - KCscCoeff + ImplicitSourceTerm(KCspCoeff)

>>> eqs = ((KMVar, KMEq), (TMVar, TMEq), (TCVar, TCEq), (P3Var, P3Eq), (P2Var, P2Eq), (KCVar, KCEq))

>>> if __name__ == '__main__':
...     steps = 100
... else:
...     steps = 28

>>> for i in range(steps):
...     for var, eqn in eqs:
...         var.updateOld()
...     for var, eqn in eqs:
...         eqn.solve(var, dt=1.)

>>> accuracy = 1e-2
>>> print(KMVar.allclose(params['KM'], atol=accuracy))
1
>>> print(TMVar.allclose(params['TM'], atol=accuracy))
1
>>> print(TCVar.allclose(params['TC'], atol=accuracy))
1
>>> print(P2Var.allclose(params['P2'], atol=accuracy))
1
>>> print(P3Var.allclose(params['P3'], atol=accuracy))
1
>>> print(KCVar.allclose(params['KC'], atol=accuracy))
1

>>> PNView = PN / PN.cellVolumeAverage
>>> PNView.name = 'PN'

>>> KMView = KMVar / KMVar.cellVolumeAverage
>>> KMView.name = 'KM'

>>> TMView = TMVar / TMVar.cellVolumeAverage
>>> TMView.naem = 'TM'

>>> RVar[:] = params['S'] + (1 + params['S']) * params['G'] * numerix.cos((2 * numerix.pi * mesh.cellCenters[0]) / L)

>>> if __name__ == '__main__':
...     KMViewer = Viewer((PNView, KMView, TMView), title = 'Gradient Stimulus: Profile')
...
...     for i in range(100):
...         for var, eqn in eqs:
...             var.updateOld()
...         for var, eqn in eqs:
...             eqn.solve(var, dt=0.1)
...
...         KMViewer.plot()
...
...     input("finished")
"""

__docformat__ = 'restructuredtext'

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())



"""

This example advects a 2 by 2 initially square region outwards.
The example checks for global conservation of surfactant.

Advect the interface and check the position.

   >>> distanceVariable.calcDistanceFunction() #doctest: +LSM
   >>> initialSurfactant = numerix.sum(surfactantVariable)
   >>> for step in range(steps):
   ...     distanceVariable.updateOld()
   ...     surfactantEquation.solve(surfactantVariable, dt=1)
   ...     advectionEquation.solve(distanceVariable, dt = timeStepDuration) #doctest: +LSM
   >>> print(numerix.allclose(initialSurfactant, numerix.sum(surfactantVariable))) #doctest: +LSM
   1



"""
__docformat__ = 'restructuredtext'

from fipy import input
from fipy import CellVariable, SurfactantVariable, Grid2D, DistanceVariable, TransientTerm, ExplicitUpwindConvectionTerm, AdvectionTerm, Viewer
from fipy.tools import numerix

L = 1.
dx = 0.1
velocity = 1.
cfl = 0.1
distanceToTravel = L / 5.
boxSize = .2

nx = L // dx
ny = L // dx

steps = int(distanceToTravel / dx / cfl)

timeStepDuration = cfl * dx / velocity

mesh = Grid2D(dx = dx, dy = dx, nx = nx, ny = ny)

x0 = (L - boxSize) / 2
x1 = (L + boxSize) / 2

distanceVariable = DistanceVariable(
    mesh = mesh,
    value = 1.,
    hasOld = 1
    )

x, y = mesh.cellCenters
distanceVariable.setValue(-1, where=((x0 < x) & (x < x1)) & ((x0 < y) & (y < x1)))


surfactantVariable = SurfactantVariable(
    distanceVar = distanceVariable,
    value = 1.
    )


from fipy.variables.surfactantConvectionVariable import SurfactantConvectionVariable
surfactantEquation = TransientTerm() - \
    ExplicitUpwindConvectionTerm(SurfactantConvectionVariable(distanceVariable))

advectionEquation = TransientTerm() + AdvectionTerm(velocity)

if __name__ == '__main__':
    distanceViewer = Viewer(vars=distanceVariable,
                            datamin=-.001, datamax=.001)
    surfactantViewer = Viewer(vars=surfactantVariable,
                              datamin=0., datamax=2.)


    distanceVariable.calcDistanceFunction()

    for step in range(steps):
        print(numerix.sum(surfactantVariable))
        distanceVariable.updateOld()
        surfactantEquation.solve(surfactantVariable, dt=1)
        advectionEquation.solve(distanceVariable, dt = timeStepDuration)
        distanceViewer.plot()
        surfactantViewer.plot()

    surfactantEquation.solve(surfactantVariable, dt=1)

    distanceViewer.plot()
    surfactantViewer.plot()
    print(surfactantVariable)
    input('finished')



"""

This input file solves a steady-state 1D diffusion problem as in
`./examples/diffusion/mesh1D.py`. The difference being that it uses a tensor for
the diffusion coefficient, even though the coefficient is isotropic.

>>> from fipy import Grid2D, CellVariable, DiffusionTerm, Viewer

>>> Lx = 20
>>> mesh = Grid2D(nx=20, ny=20)
>>> x, y = mesh.cellCenters

>>> valueLeft = 0.
>>> valueRight = 1.

>>> var = CellVariable(name = "solution variable",
...                    mesh = mesh,
...                    value = valueLeft)

>>> var.constrain(valueLeft, mesh.facesLeft)
>>> var.constrain(valueRight, mesh.facesRight)

>>> DiffusionTerm(coeff=(((1., 0.),
...                       (0., 1.)),)).solve(var)

>>> if __name__ == '__main__':
...     viewer = Viewer(vars=var).plot()

>>> analyticalArray = valueLeft + (valueRight - valueLeft) * x / Lx
>>> print(var.allclose(analyticalArray, atol = 0.025))
1
"""
__docformat__ = 'restructuredtext'

from fipy import input

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())

    input('finished')




import json
import platform
import subprocess
import sys

__all__ = ["conda_info", "pip_info", "package_info", "platform_info", "nix_info"]

def conda_info(conda="conda"):
    """Collect information about conda environment.

    Parameters
    ----------
    conda : str
        Name of conda executable (default: "conda").

    Returns
    -------
    dict
        Result of `conda info` and `conda env export` for active conda
        environment.
    """
    info = {}
    p = subprocess.Popen([conda, "info", "--json"], stdout=subprocess.PIPE)
    stdout, _ = p.communicate()
    stdout = stdout.decode('ascii')

    info["conda_info"] = json.loads(stdout)

    p = subprocess.Popen([conda, "env", "export",
                          "--prefix", info["conda_info"]["active_prefix"],
                          "--json"],
                         stdout=subprocess.PIPE)
    stdout, _ = p.communicate()
    stdout = stdout.decode('ascii')
    info["conda_env"] = json.loads(stdout)

    return info

def pip_info(python="python"):
    """Collect information about pip environment.

    Parameters
    ----------
    python : str
        Name of Python executable (default: "python").

    Returns
    -------
    list of dict
        Result of `pip list --format json`.
    """
    info = {}
    p = subprocess.Popen([python, "-m", "pip", "list", "--format", "json"], stdout=subprocess.PIPE)
    stdout, _ = p.communicate()
    stdout = stdout.decode('ascii')

    info["pip"] = json.loads(stdout)

    return info

def nix_info():
    """Collect information about nix environment.

    Returns
    -------
    dict
        Result of::

            nix derivation show .#fipy
    """
    info = {}

    # Better output would be from
    #
    #     $ nix-store -q --tree $(nix-store --realize $(nix eval --raw .#fipy.drvPath))
    #
    # However, this is difficult to execute and can require another build.
    # Also doesn't return json.

    p = subprocess.Popen(
        [
         "nix",
         "derivation",
         "show",
         ".#fipy"
        ], stdout=subprocess.PIPE)

    stdout, _ = p.communicate()
    stdout = stdout.decode('ascii')

    info["nix"] = json.loads(stdout)

    return info

def package_info():
    """Collect information about installed packages FiPy uses.

    Returns
    -------
    dict
        Versions of important Python packages.
    """
    packages = {}

    packages['python'] = sys.version #.replace('\n', '| ')

    for pkg in ['fipy', 'numpy', 'scipy', 'matplotlib', 'mpi4py', 'petsc4py', 'pyamgx']:
        try:
            mod = __import__(pkg)

            packages[pkg] = mod.__version__
        except ImportError as e:
            packages[pkg] = 'not installed'

        except Exception as e:
            packages[pkg] = 'version check failed: {}'.format(e)

    ## PyTrilinos
    try:
        import PyTrilinos
        packages['PyTrilinos'] = PyTrilinos.version()
    except ImportError as e:
        packages['PyTrilinos'] = 'not installed'
    except Exception as e:
        packages['PyTrilinos'] = 'version check failed: {}'.format(e)

    ## Mayavi uses a non-standard approach for storing its version number.
    try:
        from mayavi.__version__ import __version__ as mayaviversion
        packages['mayavi'] = mayaviversion
    except ImportError as e:
        try:
            from enthought.mayavi.__version__ import __version__ as mayaviversion
            packages['mayavi'] = mayaviversion
        except ImportError as e:
            packages['mayavi'] = 'not installed'
    except Exception as e:
        packages['mayavi'] = 'version check failed: {}'.format(e)

    ## Gmsh version
    try:
        from fipy.meshes.gmshMesh import gmshVersion
        gmshversion = gmshVersion()
        if gmshversion is None:
            packages['gmsh'] = 'not installed'
        else:
            packages['gmsh'] = gmshversion
    except Exception as e:
        packages['gmsh'] = 'version check failed: {}'.format(e)

    try:
        from fipy.solvers import solver_suite
        packages['solver'] = solver_suite
    except Exception as e:
        packages['solver'] = str(e)

    return packages

def platform_info():
    """Collect information about platform Python is running in.

    Returns
    -------
    dict
        Data extracted from `platform` package.
    """
    return {
        "architecture": platform.architecture(),
        "machine": platform.machine(),
        "node": platform.node(),
        "platform": platform.platform(),
        "processor": platform.processor(),
        "release": platform.release(),
        "system": platform.system(),
        "version": platform.version()
    }



__all__ = []

class _Vertex(object):
    def __init__(self, ID, x, y):
        self.ID = ID
        self.up = None
        self.down = None
        self.closeVertices = None
        self.inLine = False
        self.x = x
        self.y = y

    def getX(self):
        return self.x

    def getY(self):
        return self.y

    def distance(self, vertex):
        from fipy.tools import numerix
        return numerix.sqrt((vertex.getX() - self.getX())**2 + (vertex.getY() - self.getY())**2)

    def getID(self):
        return self.ID

    def setCloseVertices(self, vertices):
        self.closeVertices = vertices

    def getCloseVertices(self):
        return self.closeVertices

    def getUp(self):
        return self.up

    def getDown(self):
        return self.down

    def setInLineTrue(self):
        self.inLine = True

    def getInLine(self):
        return self.inLine

    def setUp(self, vertex):
        self.setInLineTrue()
        self.up = vertex

    def setDown(self, vertex):
        self.setInLineTrue()
        self.down = vertex

class _Line(object):
    def __init__(self, seedVertex):
        if seedVertex.getUp() is not None or \
           seedVertex.getDown() is not None or \
           seedVertex.getInLine():
            raise ValueError('bad seedVertex')

        setVs = []

        vertex = seedVertex
        while vertex is not None and vertex.getUp() is None:
            for v in vertex.getCloseVertices():
                if v not in setVs:
                    if vertex in v.getCloseVertices():
                        if v.getDown() is None:
                            if vertex.getUp() is None:
                                if vertex.getDown() is not v:
                                    vertex.setUp(v)
                                    v.setDown(vertex)

            setVs.append(vertex)
            vertex = vertex.getUp()



        self.startVertex = seedVertex
        vertex = seedVertex
        while vertex is not None and vertex.getDown() is None:
            for v in vertex.getCloseVertices():
                if v not in setVs:
                    if vertex in v.getCloseVertices():
                        if v.getUp() is None:
                            if vertex.getDown() is None:
                                if vertex.getUp() is not v:
                                    vertex.setDown(v)
                                    v.setUp(vertex)

            self.startVertex = vertex
            setVs.append(vertex)
            vertex = vertex.getDown()

        self.startVertex.setInLineTrue()

    def getVertexListIDs(self):
        vertexListIDs = []
        vertexListIDs.append(self.startVertex.getID())
        v = self.startVertex.getUp()
        while v is not None and v is not self.startVertex:
            vertexListIDs.append(v.getID())
            v = v.getUp()

        if v is not None:
            vertexListIDs.append(v.getID())
        return vertexListIDs

def _getOrderedLines(IDs, coordinates, thresholdDistance = 0.0):
    """
    This function takes a set of IDs and corresponding coordinates and makes
    a set of closed curves.

    The following are a general set of test cases.

       >>> _getOrderedLines((0, 1, 2, 3), ((0, 0), (2, 0), (0, 1), (2, 1)))
       [[0, 2, 3, 1]]
       >>> _getOrderedLines((0, 1, 2, 3, 4), ((-10, -10), (0, 0), (2, 0), (0, 1), (2, 1)))
       [[0], [1, 3, 4, 2]]
       >>> _getOrderedLines((0, 1, 2, 3), ((0, 0), (0.9, 0), (0, 1), (1, 1)))
       [[0, 1, 3, 2]]
       >>> _getOrderedLines((0, 1, 2, 3, 4, 5), ((0, 0), (1, 0), (2, 0), (0, 1.1), (1, 1.1), (2, 1.1)))
       [[0, 1, 2, 5, 4, 3]]
       >>> _getOrderedLines((4, 5, 0, 1, 3, 2, 6), ((0, 0), (1, 0), (3, 0), (0, 1.1), (1, 1.1), (3, 1), (4, 1)))
       [[4, 5, 3, 1], [0, 2, 6]]
       >>> _getOrderedLines((4, 5, 3, 2, 1, 0, 9, 8, 7, 6), ((0, 0), (1, 0), (0, 1.1), (1, 1.1), (0, 3), (1, 3), (-1, 4), (2, 4), (-2, 4), (3, 4)))
       [[4, 5, 2, 3], [7, 9, 1, 0, 8, 6]]
       >>> _getOrderedLines(list(range(7)), ((-7, 0), (-6, 0), (-5, 0), (0, 0), (5, 0), (6, 0), (7, 0)))
       [[0, 1, 2], [3], [4, 5, 6]]
       >>> _getOrderedLines(list(range(7)), ((-7, 0), (-6, 0), (-5, 0), (0, 0), (5, 0), (6, 0), (7, 0)), thresholdDistance = 5.5)
       [[0, 1, 2, 3, 4, 5, 6]]

    Parameters
    ----------
    IDs : ndarray of int
    coordinates : ndarray
        An array of coordinates of the same length as IDs.
    """

    from fipy.tools import numerix
    coordinates = numerix.array(coordinates)
    closeIDs = numerix.zeros((len(IDs), len(IDs)), 'l')
    vertices = []
    for ID in IDs:
        distances = numerix.zeros(len(IDs), 'd')
        for i in range(len(coordinates[0,:])):
            distances += (coordinates[:, i] - coordinates[ID, i])**2
        vertices.append(_Vertex(ID, coordinates[ID, 0], coordinates[ID, 1]))
        closeIDs[ID,:] = numerix.argsort(distances)


    for ID in IDs:
        i = 1
        closeVertices = []
        while i < 3 or vertices[ID].distance(vertices[closeIDs[ID, i]]) < thresholdDistance:
            closeVertices.append(vertices[closeIDs[ID, i]])
            i += 1

        vertices[ID].setCloseVertices(closeVertices)

    listOfVertexLists = []

    for vertex in vertices:
        if not vertex.getInLine():
            listOfVertexLists.append(_Line(vertex).getVertexListIDs())

    return listOfVertexLists

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()



from fipy.tools import numerix
from fipy.tools.comms.commWrapper import CommWrapper

__all__ = ["DummyComm"]

class DummyComm(CommWrapper):
    @property
    def procID(self):
        return 0
        
    @property
    def Nproc(self):
        return 1



__docformat__ = 'restructuredtext'

from PyTrilinos import ML

from .multilevelPreconditioner import MultilevelPreconditioner

__all__ = ["MultilevelSolverSmootherPreconditioner"]

class MultilevelSolverSmootherPreconditioner(MultilevelPreconditioner):
    """Multilevel preconditioner using Aztec solvers as smoothers for :class:`~fipy.solvers.trilinos.trilinosSolver.TrilinosSolver`.
    """

    @property
    def _parameterList(self):
        return {
            "output": 0,
            "max levels": self.levels,
            "smoother: type": "Aztec",
            "smoother: Aztec as solver": True
        }



from .defaultPreconditioner import *
from .icPreconditioner import *
from .iluPreconditioner import *
from .hyprePreconditioner import *
from .jacobiPreconditioner import *
from .luPreconditioner import *
from .ssorPreconditioner import *

__all__ = []
__all__.extend(defaultPreconditioner.__all__)
__all__.extend(icPreconditioner.__all__)
__all__.extend(iluPreconditioner.__all__)
__all__.extend(hyprePreconditioner.__all__)
__all__.extend(jacobiPreconditioner.__all__)
__all__.extend(luPreconditioner.__all__)
__all__.extend(ssorPreconditioner.__all__)



__docformat__ = 'restructuredtext'

from fipy.solvers.scipy.scipyKrylovSolver import ScipyKrylovSolver
from scipy.sparse.linalg import bicgstab

__all__ = ["LinearBicgstabSolver"]

class LinearBicgstabSolver(ScipyKrylovSolver):
    """Interface to the Biconjugate Gradient (Stabilized) (:term:`BiCGSTAB`)
    solver in :ref:`Scipy`.

    No preconditioning by default.
    """

    solveFnc = staticmethod(bicgstab)



__docformat__ = 'restructuredtext'

from fipy.solvers.preconditioner import SolverModifyingPreconditioner

__all__ = ["PETScPreconditioner"]

class PETScPreconditioner(SolverModifyingPreconditioner):
    """Base class preconditioners of  for :class:`~fipy.solvers.petsc.petscSolver.PETScSolver`.

    .. attention:: This class is abstract. Always create one of its subclasses.
    """

    def _applyToSolver(self, solver, matrix):
        solver.getPC().setType(self.pctype)



"""Test boundary conditions
"""

__all__ = []

from fipy.tests.doctestPlus import _LateImportDocTestSuite
import fipy.tests.testProgram

def _suite():
    return _LateImportDocTestSuite(
        docTestModuleNames = (
            'fipy.boundaryConditions.boundaryCondition',
            'fipy.boundaryConditions.fixedFlux',
        ))

if __name__ == '__main__':
    fipy.tests.testProgram.main(defaultTest='_suite')



__docformat__ = 'restructuredtext'

from fipy.viewers.viewer import AbstractViewer
from fipy.viewers import MeshDimensionError
from fipy.tools import numerix

class MayaviSurfactantViewer(AbstractViewer):

    """
    The `MayaviSurfactantViewer` creates a viewer with the Mayavi_ python
    plotting package that displays a `DistanceVariable`.

    .. _Mayavi: http://mayavi.sourceforge.net/

    """

    def __init__(self, distanceVar, surfactantVar=None, levelSetValue=0., title=None, smooth=0, zoomFactor=1., animate=False, limits={}, **kwlimits):
        """
        Create a `MayaviSurfactantViewer`.

            >>> from fipy import *
            >>> dx = 1.
            >>> dy = 1.
            >>> nx = 11
            >>> ny = 11
            >>> Lx = ny * dy
            >>> Ly = nx * dx
            >>> mesh = Grid2D(dx = dx, dy = dy, nx = nx, ny = ny)
            >>> # from fipy.models.levelSet.distanceFunction.distanceVariable import DistanceVariable
            >>> var = DistanceVariable(mesh = mesh, value = -1)

            >>> x, y = mesh.cellCenters

            >>> var.setValue(1, where=(x - Lx / 2.)**2 + (y - Ly / 2.)**2 < (Lx / 4.)**2)
            >>> var.calcDistanceFunction()
            >>> viewer = MayaviSurfactantViewer(var, smooth = 2)
            >>> viewer.plot()
            >>> viewer._promptForOpinion()
            >>> del viewer

            >>> var = DistanceVariable(mesh = mesh, value = -1)

            >>> var.setValue(1, where=(y > 2. * Ly / 3.) | ((x > Lx / 2.) & (y > Ly / 3.)) | ((y < Ly / 6.) & (x > Lx / 2)))
            >>> var.calcDistanceFunction()
            >>> viewer = MayaviSurfactantViewer(var)
            >>> viewer.plot()
            >>> viewer._promptForOpinion()
            >>> del viewer

            >>> viewer = MayaviSurfactantViewer(var, smooth = 2)
            >>> viewer.plot()
            >>> viewer._promptForOpinion()
            >>> del viewer

        Parameters
        ----------
        distanceVar : ~fipy.variables.distanceVariable.DistanceVariable
        levelSetValue : float
            the value of the contour to be displayed
        title : str
            displayed at the top of the `Viewer` window
        animate : bool
            whether to show only the initial condition and the
            moving top boundary or to show all contours (Default)
        limits : dict, optional
            a (deprecated) alternative to limit keyword arguments
        xmin, xmax, ymin, ymax, zmin, zmax, datamin, datamax : float, optional
            displayed range of data. A 1D `Viewer` will only use `xmin` and
            `xmax`, a 2D viewer will also use `ymin` and `ymax`, and so on. All
            viewers will use `datamin` and `datamax`. Any limit set to a
            (default) value of `None` will autoscale.
        """

        kwlimits.update(limits)
        AbstractViewer.__init__(self, vars=[], title=title, **kwlimits)
        import mayavi
        self._viewer = mayavi.mayavi()
        self.distanceVar = distanceVar
        if surfactantVar is None:
            self.surfactantVar = numerix.zeros(len(self.distanceVar), 'd')
        else:
            self.surfactantVar = surfactantVar
        self.smooth = smooth
        self.zoomFactor = zoomFactor

        self.animate = animate
        if animate:
            self._initialCondition = None

        if distanceVar.mesh.dim != 2:
            raise MeshDimensionError('The MayaviIsoViewer only works for 2D meshes.')

    def _getStructure(self):

        ##maxX = self.distanceVar.mesh.faceCenters[0].max()
        ##minX = self.distanceVar.mesh.faceCenters[0].min()

        IDs = numerix.nonzero(self.distanceVar._cellInterfaceFlag)[0]
        coordinates = numerix.take(numerix.array(self.distanceVar.mesh.cellCenters).swapaxes(0, 1), IDs)

        coordinates -= numerix.take(numerix.array(self.distanceVar.grad * self.distanceVar).swapaxes(0, 1), IDs)

        coordinates *= self.zoomFactor

        shiftedCoords = coordinates.copy()
        shiftedCoords[:, 0] = -coordinates[:, 0] ##+ (maxX - minX)
        coordinates = numerix.concatenate((coordinates, shiftedCoords))

        from .lines import _getOrderedLines

        lines = _getOrderedLines(list(range(2 * len(IDs))), coordinates, thresholdDistance = self.distanceVar.mesh._cellDistances.min() * 10)

        data = numerix.take(self.surfactantVar, IDs)

        data = numerix.concatenate((data, data))

        tmpIDs = numerix.nonzero(data > 0.0001)[0]
        if len(tmpIDs) > 0:
            val = numerix.take(data, tmpIDs).min()
        else:
            val = 0.0001

        data = numerix.where(data < 0.0001,
                             val,
                             data)

        for line in lines:
            if len(line) > 2:
                for smooth in range(self.smooth):
                    for arr in (coordinates, data):
                        tmp = numerix.take(arr, line)
                        tmp[1:-1] = tmp[2:] * 0.25 + tmp[:-2] * 0.25 + tmp[1:-1] * 0.5
                        if len(arr.shape) > 1:
                            for i in range(len(arr[0])):
                                arrI = arr[:, i].copy()
                                numerix.put(arrI, line, tmp[:, i])
                                arr[:, i] = arrI
                        else:
                            numerix.put(arrI, line, tmp)

        name = self.title
        name = name.strip()
        if name == '':
            name = None

        coords = numerix.zeros((coordinates.shape[0], 3), 'd')
        coords[:, :coordinates.shape[1]] = coordinates

        import pyvtk

        ## making lists as pyvtk doesn't know what to do with numpy arrays

        coords = list(coords)
        coords = [[float(coord[0]), float(coord[1]), float(coord[2])] for coord in coords]

        data = list(data)
        data = [float(item) for item in data]

        return (pyvtk.UnstructuredGrid(points = coords,
                                       poly_line = lines),
                pyvtk.PointData(pyvtk.Scalars(data, name = name)))

    def plot(self, filename = None):

        structure, data = self._getStructure()

        import pyvtk
        import tempfile
        import os

        if self.animate:
            if self._initialCondition is None:
                data = pyvtk.VtkData(structure, 0)
                (f, tempFileName) = tempfile.mkstemp('.vtk')
                data.tofile(tempFileName)
                self._viewer.open_vtk(tempFileName, config=0)
                os.close(f)
                os.remove(tempFileName)

                self._viewer.load_module('SurfaceMap', 0)

                rw = self._viewer.get_render_window()
                rw.z_plus_view()

                self._initialCondition = self._viewer.get_current_dvm_name()
            else:
                self._viewer.mayavi.del_dvm(self._viewer.get_current_dvm_name())

        data = pyvtk.VtkData(structure, data)

        (f, tempFileName) = tempfile.mkstemp('.vtk')
        data.tofile(tempFileName)
        self._viewer.open_vtk(tempFileName, config=0)

        os.close(f)
        os.remove(tempFileName)
        self._viewer.load_module('SurfaceMap', 0)

        if not self.animate:
            rw = self._viewer.get_render_window()
            rw.z_plus_view()

        ## display legend
        dvm = self._viewer.get_current_dvm()
        mm = dvm.get_current_module_mgr()
        slh = mm.get_scalar_lut_handler()
        slh.legend_on.set(1)
        slh.legend_on_off()

        ## display legend with correct range
        slh.range_on_var.set(1)
        slh.v_range_on_var.set(1)

        xmax = self._getLimit('datamax', default=self.surfactantVar.max())
        xmin = self._getLimit('datamin', default=self.surfactantVar.min())

        slh.range_var.set((xmin, xmax))
        slh.set_range_var()

        slh.v_range_var.set((float(self.surfactantVar.min()), float(self.surfactantVar.max())))
        slh.set_v_range_var()

        self._viewer.Render()

        if filename is not None:
            self._viewer.renwin.save_png(filename)

if __name__ == "__main__":
    import fipy.tests.doctestPlus
    fipy.tests.doctestPlus.execButNoTest()



r"""Solve a coupled set of diffusion equations in two dimensions.

This example solves a diffusion problem and demonstrates the use of
applying boundary condition patches.

.. index::
   single: Grid2D

>>> from fipy import CellVariable, Grid2D, Viewer, TransientTerm, DiffusionTerm
>>> from fipy.tools import numerix

>>> nx = 20
>>> ny = nx
>>> dx = 1.
>>> dy = dx
>>> L = dx * nx
>>> mesh = Grid2D(dx=dx, dy=dy, nx=nx, ny=ny)

We create a :class:`~fipy.variables.cellVariable.CellVariable` and initialize it to zero:

>>> phi = CellVariable(name = "solution variable",
...                    mesh = mesh,
...                    value = 0.)

and then create a diffusion equation.  This is solved by default with an
iterative conjugate gradient solver.

>>> D = 1.
>>> eq = TransientTerm(var=phi) == DiffusionTerm(coeff=D, var=phi)

We apply Dirichlet boundary conditions

>>> valueTopLeft = 0
>>> valueBottomRight = 1

to the top-left and bottom-right corners.  Neumann boundary conditions
are automatically applied to the top-right and bottom-left corners.

>>> x, y = mesh.faceCenters
>>> facesTopLeft = ((mesh.facesLeft & (y > L / 2))
...                 | (mesh.facesTop & (x < L / 2)))
>>> facesBottomRight = ((mesh.facesRight & (y < L / 2))
...                     | (mesh.facesBottom & (x > L / 2)))

>>> phi.constrain(valueTopLeft, facesTopLeft)
>>> phi.constrain(valueBottomRight, facesBottomRight)

We create a viewer to see the results

.. index::
   pair: module; fipy.viewers

>>> if __name__ == '__main__':
...     viewer = Viewer(vars=phi, datamin=0., datamax=1.)
...     viewer.plot()

and solve the equation by repeatedly looping in time:

>>> timeStepDuration = 10 * 0.9 * dx**2 / (2 * D)
>>> steps = 10
>>> for step in range(steps):
...     eq.solve(dt=timeStepDuration)
...     if __name__ == '__main__':
...         viewer.plot()

.. image:: /figures/examples/diffusion/mesh20x20transient.*
   :width: 90%
   :align: center

We can test the value of the bottom-right corner cell.

>>> print(numerix.allclose(phi(((L,), (0,))), valueBottomRight, atol = 1e-2))
1

>>> from fipy import input
>>> if __name__ == '__main__':
...     input("Implicit transient diffusion. Press <return> to proceed...")

----

We can also solve the steady-state problem directly

>>> DiffusionTerm(var=phi).solve()
>>> if __name__ == '__main__':
...     viewer.plot()

.. image:: /figures/examples/diffusion/mesh20x20steadyState.*
   :width: 90%
   :align: center

and test the value of the bottom-right corner cell.

>>> print(numerix.allclose(phi(((L,), (0,))), valueBottomRight, atol = 1e-2))
1

>>> from fipy import input
>>> if __name__ == '__main__':
...     input("Implicit steady-state diffusion. Press <return> to proceed...")
"""

__docformat__ = 'restructuredtext'

##from fipy.tools.profiler.profiler import Profiler
##from fipy.tools.profiler.profiler import calibrate_profiler

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())



""" Solve a phase-field evolution and diffusion of four species in one-dimension.

The same procedure used to construct the two-component phase field
diffusion problem in :mod:`examples.phase.binary` can be used to build up a
system of multiple components. Once again, we'll focus on 1D.

.. index::
   single: Grid1D

>>> from fipy import CellVariable, Grid1D, TransientTerm, DiffusionTerm, ImplicitSourceTerm, PowerLawConvectionTerm, DefaultAsymmetricSolver, Viewer
>>> from fipy.tools import numerix

>>> nx = 400
>>> dx = 0.01
>>> L = nx * dx
>>> mesh = Grid1D(dx = dx, nx = nx)

We consider a free energy density :math:`f(\\phi, C_0,\\ldots,C_N, T)`
that is a function of phase :math:`\\phi`

.. index::
   single: CellVariable

>>> phase = CellVariable(mesh=mesh, name='phase', value=1., hasOld=1)

interstitial components :math:`C_0 \\ldots C_M`

>>> interstitials = [
...     CellVariable(mesh=mesh, name='C0', hasOld=1)
... ]

substitutional components :math:`C_{M+1} \\ldots C_{N-1}`

>>> substitutionals = [
...     CellVariable(mesh=mesh, name='C1', hasOld=1),
...     CellVariable(mesh=mesh, name='C2', hasOld=1),
... ]

a "solvent" :math:`C_N` that is constrained by the concentrations of the
other substitutional species, such that :math:`C_N = 1 - \\sum_{j=M}^{N-1}
C_j`,

>>> solvent = 1
>>> for Cj in substitutionals:
...     solvent -= Cj
>>> solvent.name = 'CN'

and temperature :math:`T`

>>> T = 1000

The free energy density of such a system can be written as

.. math::

   f(\\phi, C_0, \\ldots, C_N, T)
   &= \\sum_{j=0}^N C_j \\left[ \\mu^\\circ_j(\\phi, T) + R T \\ln \\frac{C_j}{\\rho} \\right]

where

>>> R = 8.314 # J / (mol K)

is the gas constant. As in the binary case,

.. math::

   \\mu^\\circ_j(\\phi, T) = p(\\phi) \\mu_j^{\\circ S}(T)
   + \\left(1 - p(\\phi)\\right) \\mu_j^{\\circ L}(T) + \\frac{W_j}{2} g(\\phi)

is constructed with the free energies of the pure components in each
phase, given the "tilting" function

>>> def p(phi):
...     return phi**3 * (6 * phi**2 - 15 * phi + 10)

and the "double well" function

>>> def g(phi):
...     return (phi * (1 - phi))**2

We consider a very simplified model that has partial molar volumes
:math:`\\bar{V}_0 = \\cdots = \\bar{V}_{M} = 0` for the "interstitials" and
:math:`\\bar{V}_{M+1} = \\cdots = \\bar{V}_{N} = 1` for the "substitutionals".
This approximation has been used in a number of models where density
effects are ignored, including the treatment of electrons in
electrodeposition processes :cite:`ElPhFI` :cite:`ElPhFII`. Under these
constraints

.. math::

   \\frac{\\partial f}{\\partial \\phi}
   &= \\sum_{j=0}^N C_j \\frac{\\partial f_j}{\\partial \\phi}
   \\nonumber \\\\
   &= \\sum_{j=0}^N C_j \\left[
       \\mu_j^{\\circ SL}(T) p'(\\phi) + \\frac{W_j}{2} g'(\\phi)
   \\right]
   \\\\
   \\frac{\\partial f}{\\partial C_j}
   &= \\left[\\mu^\\circ_j(\\phi, T) + R T \\ln \\frac{C_j}{\\rho} \\right]
   \\nonumber \\\\
   &= \\mu_j(\\phi, C_j , T)
   \\qquad\\text{for \\( j = 0\\ldots M \\)}

and

.. math::

   \\frac{\\partial f}{\\partial C_j}
   &= \\left[\\mu^\\circ_j(\\phi, T) + R T \\ln \\frac{C_j}{\\rho} \\right]
   - \\left[\\mu^\\circ_N(\\phi, T) + R T \\ln \\frac{C_N}{\\rho} \\right]
   \\nonumber \\\\
   &= \\left[\\mu_j(\\phi, C_j, T) - \\mu_N(\\phi, C_N, T) \\right]
   \\qquad\\text{for \\( j = M+1\\ldots N-1 \\)}

where :math:`\\mu_j^{\\circ SL}(T) \\equiv \\mu_j^{\\circ S}(T) - \\mu_j^{\\circ
L}(T)` and where :math:`\\mu_j` is the classical chemical potential of
component :math:`j` for the binary species and :math:`\\rho = 1 +
\\sum_{j=0}^{M} C_j` is the total molar density.

>>> rho = 1.
>>> for Cj in interstitials:
...     rho += Cj

:math:`p'(\\phi)` and :math:`g'(\\phi)` are the partial derivatives of of :math:`p`
and :math:`g` with respect to :math:`\\phi`

>>> def pPrime(phi):
...     return 30. * g(phi)

>>> def gPrime(phi):
...     return 2. * phi * (1 - phi) * (1 - 2 * phi)

We "cook" the standard potentials to give the desired solid and liquid
concentrations, with a solid phase rich in interstitials and the solvent
and a liquid phase rich in the two substitutional species.

>>> interstitials[0].S = 0.3
>>> interstitials[0].L = 0.4
>>> substitutionals[0].S = 0.4
>>> substitutionals[0].L = 0.3
>>> substitutionals[1].S = 0.2
>>> substitutionals[1].L = 0.1
>>> solvent.S = 1.
>>> solvent.L = 1.
>>> for Cj in substitutionals:
...     solvent.S -= Cj.S
...     solvent.L -= Cj.L

>>> rhoS = rhoL = 1.
>>> for Cj in interstitials:
...     rhoS += Cj.S
...     rhoL += Cj.L

.. index::
   single: log

>>> for Cj in interstitials + substitutionals + [solvent]:
...     Cj.standardPotential = R * T * (numerix.log(Cj.L/rhoL)
...                                     - numerix.log(Cj.S/rhoS))

>>> for Cj in interstitials:
...     Cj.diffusivity = 1.
...     Cj.barrier = 0.

>>> for Cj in substitutionals:
...     Cj.diffusivity = 1.
...     Cj.barrier = R * T

>>> solvent.barrier = R * T

----

We create the phase equation

.. math::

   \\frac{1}{M_\\phi}\\frac{\\partial \\phi}{\\partial t}
   = \\kappa_\\phi \\nabla^2 \\phi
   - \\sum_{j=0}^N C_j \\left[
              \\mu_j^{\\circ SL}(T) p'(\\phi) + \\frac{W_j}{2} g'(\\phi)
          \\right]

with a semi-implicit source just as in :mod:`examples.phase.simple` and
:mod:`examples.phase.binary`

>>> enthalpy = 0.
>>> barrier = 0.
>>> for Cj in interstitials + substitutionals + [solvent]:
...     enthalpy += Cj * Cj.standardPotential
...     barrier += Cj * Cj.barrier

>>> mPhi = -((1 - 2 * phase) * barrier + 30 * phase * (1 - phase) * enthalpy)
>>> dmPhidPhi = 2 * barrier - 30 * (1 - 2 * phase) * enthalpy
>>> S1 = dmPhidPhi * phase * (1 - phase) + mPhi * (1 - 2 * phase)
>>> S0 = mPhi * phase * (1 - phase) - S1 * phase

>>> phase.mobility = 1.
>>> phase.gradientEnergy = 25
>>> phase.equation = TransientTerm(coeff=1/phase.mobility) \\
...   == DiffusionTerm(coeff=phase.gradientEnergy) \\
...      + S0 + ImplicitSourceTerm(coeff = S1)

We could construct the diffusion equations one-by-one, in the manner of
:mod:`examples.phase.binary`, but it is better to take advantage of the full
scripting power of the Python language, where we can easily loop over
components or even make "factory" functions if we desire. For the
interstitial diffusion equations, we arrange in canonical form as before:

.. math::

   \\underbrace{
       \\frac{\\partial C_j}{\\partial t}
       \\vphantom{\\left\\{
           \\overbrace{
               \\left[
                   \\mu_j^{\\circ SL} \\nabla p(\\phi)
               \\right]
           }^{\\text{phase transformation}}
       \\right\\}}
   }_{\\text{transient}}
   &= \\underbrace{
       D_j\\nabla^2 C_j
       \\vphantom{\\left\\{
           \\overbrace{
               \\left[
                   \\mu_j^{\\circ SL} \\nabla p(\\phi)
               \\right]
           }^{\\text{phase transformation}}
       \\right\\}}
   }_{\\text{diffusion}} \\\\
   & \\qquad + \\underbrace{
       D_j\\nabla\\cdot
       \\frac{C_j}{1 + \\sum_{\\substack{k=0\\\\ k \\neq j}}^{M} C_k}
       \\left\\{
           \\overbrace{
               \\frac{\\rho}{R T}
               \\left[
                   \\mu_j^{\\circ SL} \\nabla p(\\phi)
                   + \\frac{W_j}{2} \\nabla g(\\phi)
               \\right]
           }^{\\text{phase transformation}}
           -
           \\overbrace{
               \\sum_{\\substack{i=0\\\\ i \\neq j}}^{M} \\nabla C_i
           }^{\\text{counter diffusion}}
       \\right\\}
   }_{\\text{convection}}


.. index::
   single: PowerLawConvectionTerm

>>> for Cj in interstitials:
...     phaseTransformation = (rho.harmonicFaceValue / (R * T)) \\
...       * (Cj.standardPotential * p(phase).faceGrad
...          + 0.5 * Cj.barrier * g(phase).faceGrad)
... 
...     CkSum = CellVariable(mesh=mesh, value=0.)
...     for Ck in [Ck for Ck in interstitials if Ck is not Cj]:
...         CkSum += Ck
... 
...     counterDiffusion = CkSum.faceGrad
... 
...     convectionCoeff = counterDiffusion + phaseTransformation
...     convectionCoeff *= (Cj.diffusivity
...                         / (1. + CkSum.harmonicFaceValue))
... 
...     Cj.equation = (TransientTerm()
...                    == DiffusionTerm(coeff=Cj.diffusivity)
...                    + PowerLawConvectionTerm(coeff=convectionCoeff))

----

The canonical form of the substitutional diffusion equations is

.. math::

   \\underbrace{
        \\frac{\\partial C_j}{\\partial t}
   }_{\\text{transient}}
    &= \\underbrace{
        D_{j}\\nabla^2 C_j
        \\vphantom{\\frac{\\partial C_j}{\\partial t}}
    }_{\\text{diffusion}} \\\\
    & \\qquad + \\underbrace{
        D_{j}\\nabla\\cdot
        \\frac{C_j}{1 - \\sum_{\\substack{k=M+1\\\\ k \\neq j}}^{N-1} C_k}
        \\left\\{
           \\overbrace{
                \\frac{C_N}{R T}
                \\left[
                    \\left(\\mu_j^{\\circ SL} - \\mu_N^{\\circ SL}\\right) \\nabla p(\\phi)
                    + \\frac{W_j - W_N}{2} \\nabla g(\\phi)
                \\right]
                \\vphantom{\\sum_{\\substack{i=M+1\\\\ i \\neq j}}^{N-1} \\nabla C_i}
           }^{\\text{phase transformation}}
           +
           \\overbrace{
               \\sum_{\\substack{i=M+1\\\\ i \\neq j}}^{N-1} \\nabla C_i
           }^{\\text{counter diffusion}}
        \\right\\}
    }_{\\text{convection}}

>>> for Cj in substitutionals:
...     phaseTransformation = (solvent.harmonicFaceValue / (R * T)) \\
...       * ((Cj.standardPotential - solvent.standardPotential) * p(phase).faceGrad
...          + 0.5 * (Cj.barrier - solvent.barrier) * g(phase).faceGrad)
... 
...     CkSum = CellVariable(mesh=mesh, value=0.)
...     for Ck in [Ck for Ck in substitutionals if Ck is not Cj]:
...         CkSum += Ck
... 
...     counterDiffusion = CkSum.faceGrad
... 
...     convectionCoeff = counterDiffusion + phaseTransformation
...     convectionCoeff *= (Cj.diffusivity
...                         / (1. - CkSum.harmonicFaceValue))
... 
...     Cj.equation = (TransientTerm()
...                    == DiffusionTerm(coeff=Cj.diffusivity)
...                    + PowerLawConvectionTerm(coeff=convectionCoeff))

----

We start with a sharp phase boundary

.. math::

   \\xi =
   \\begin{cases}
       1& \\text{for $x \\le L/2$,} \\\\
       0& \\text{for $x > L/2$,}
   \\end{cases}

>>> x = mesh.cellCenters[0]
>>> phase.setValue(1.)
>>> phase.setValue(0., where=x > L / 2)

and with uniform concentration fields, initially equal to the average of
the solidus and liquidus concentrations

>>> for Cj in interstitials + substitutionals:
...     Cj.setValue((Cj.S + Cj.L) / 2.)

If we're running interactively, we create a viewer

.. index::
   pair: module; viewers

>>> if __name__ == '__main__':
...     viewer = Viewer(vars=([phase]
...                           + interstitials + substitutionals
...                           + [solvent]),
...                     datamin=0, datamax=1)
...     viewer.plot()

and again iterate to equilibrium.  The initial residual is much larger than
the norm of the right-hand-side vector, so we use `"initial"` tolerance
scaling.

.. index::
   single: DefaultAsymmetricSolver

>>> solver = DefaultAsymmetricSolver(criterion="initial", tolerance=1e-8)


>>> dt = 10000
>>> for i in range(5):
...     for field in [phase] + substitutionals + interstitials:
...         field.updateOld()
...     phase.equation.solve(var = phase, dt = dt)
...     for field in substitutionals + interstitials:
...         field.equation.solve(var = field,
...                              dt = dt,
...                              solver = solver)
...     if __name__ == '__main__':
...         viewer.plot()

.. image:: /figures/examples/phase/quaternary.*
   :width: 90%
   :align: center
   :alt: phase and four composition fields in equilibrium

We can confirm that the far-field phases have remained separated

.. index::
   single: allclose

>>> X = mesh.faceCenters[0]
>>> print(numerix.allclose(phase.faceValue[X.value==0], 1.0, rtol = 1e-5, atol = 1e-5))
True
>>> print(numerix.allclose(phase.faceValue[X.value==L], 0.0, rtol = 1e-5, atol = 1e-5))
True

and that the concentration fields have appropriately segregated into
their equilibrium values in each phase

>>> equilibrium = True
>>> for Cj in interstitials + substitutionals:
...     equilibrium &= numerix.allclose(Cj.faceValue[X.value==0], Cj.S, rtol = 3e-3, atol = 3e-3).value
...     equilibrium &= numerix.allclose(Cj.faceValue[X.value==L], Cj.L, rtol = 3e-3, atol = 3e-3).value
>>> print(equilibrium)
True

.. .. bibmissing:: /refs.bib
    :sort:
"""


__docformat__ = 'restructuredtext'

from fipy import input

if __name__ == "__main__":
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())
    input("finished")

## if __name__ == '__main__':
##     ## from fipy.tools.profiler.profiler import Profiler
##     ## from fipy.tools.profiler.profiler import calibrate_profiler
##
##     # fudge = calibrate_profiler(10000)
##     # profile = Profiler('profile', fudge=fudge)
##
##     import fipy.tests.doctestPlus
##     exec(fipy.tests.doctestPlus.getScript())
##
##     # profile.stop()
##
##     raw_input("finished")
##



"""Collections of values supporting lazy evaluation
"""
__docformat__ = 'restructuredtext'

from fipy.variables.variable import *
from fipy.variables.cellVariable import *
from fipy.variables.faceVariable import *
from fipy.variables.scharfetterGummelFaceVariable import *
from fipy.variables.modularVariable import *
from fipy.variables.betaNoiseVariable import *
from fipy.variables.exponentialNoiseVariable import *
from fipy.variables.gammaNoiseVariable import *
from fipy.variables.gaussianNoiseVariable import *
from fipy.variables.uniformNoiseVariable import *
from fipy.variables.histogramVariable import *
from fipy.variables.surfactantVariable import *
from fipy.variables.surfactantConvectionVariable import *
from fipy.variables.distanceVariable import *



from fipy.tests.doctestPlus import _LateImportDocTestSuite
import fipy.tests.testProgram

def _suite():
    return _LateImportDocTestSuite(testModuleNames = (
                                       'impingement.test',
                                       'missOrientation.test',
                                   ),
                                   docTestModuleNames = (
                                       'binary',
                                       'anisotropyOLD',
                                       'anisotropy',
                                       'quaternary',
                                       'simple',
                                       'symmetry',
                                       'binaryCoupled',
                                       'polyxtal',
                                       'polyxtalCoupled'
                                   ),
                                   base = __name__)

if __name__ == '__main__':
    fipy.tests.testProgram.main(defaultTest='_suite')



__docformat__ = 'restructuredtext'

from petsc4py import PETSc

from fipy.solvers.solver import Solver
from fipy.tools import numerix
from fipy.matrices.petscMatrix import _PETScMeshMatrix

class PETScSolver(Solver):

    """
    .. attention:: This class is abstract. Always create one of its subclasses.

    """

    @property
    def globalVectors(self):
        if not hasattr(self, '_globalVectors'):
            overlappingVector = self.matrix._fipy2petscGhost(var=self.var)

            from fipy.variables.coupledCellVariable import _CoupledCellVariable
            if isinstance(self.RHSvector, _CoupledCellVariable):
                RHSvector = self.RHSvector
            else:
                RHSvector = numerix.reshape(numerix.asarray(self.RHSvector), self.var.shape)
                
            overlappingRHSvector = self.matrix._fipy2petscGhost(var=RHSvector)

            self._globalVectors = (overlappingVector, overlappingRHSvector)

        return self._globalVectors

    def _deleteGlobalVectors(self):
        if hasattr(self, "_globalVectors"):
            overlappingVector, overlappingRHSvector = self.globalVectors
            overlappingVector.destroy()
            overlappingRHSvector.destroy()
        del self._globalVectors
        
    def _rhsNorm(self, L, x, b):
        return b.norm(PETSc.NormType.NORM_2)

    def _matrixNorm(self, L, x, b):
        L.assemble()
        return L.norm(PETSc.NormType.NORM_INFINITY)

    def _residualVectorAndNorm(self, L, x, b):
        residualVector = L * x - b

        return residualVector, residualVector.norm(PETSc.NormType.NORM_2)

    @property
    def _Lxb(self):
        """Matrix, solution vector, and right-hand side vector

        Returns
        -------
        L : PETSc.Mat
            Sparse matrix
        x : PETSc.Vec
            Solution variable as ghosted vector
        b : PETSc.Vec
            Right-hand side as ghosted vector
        """
        x, b = self.globalVectors
        L = self.matrix.matrix

        if ((self.matrix == 0)
            or (L.sizes[0][1] != L.sizes[1][1])
            or (L.sizes[0][1] != x.size)):

            from fipy.terms import SolutionVariableNumberError

            raise SolutionVariableNumberError

        return (L, x, b)

    def _scatterGhosts(self, x):
        """Distribute ghost values (if any) across processes
        """
        return self.matrix._petsc2fipyGhost(vec=x)

    def _cleanup(self):
        self._deleteGlobalVectors()
        del self.var
        del self.RHSvector

    @property
    def _matrixClass(self):
        return _PETScMeshMatrix

    def _calcResidualVector(self, residualFn=None):
        if residualFn is not None:
            return residualFn(self.var, self.matrix, self.RHSvector)
        else:
            residual = self._calcResidualVector_()
            
            residual.ghostUpdate()
            with residual.localForm() as lf:
                arr = numerix.array(lf)
            residual.destroy()
            return arr

    def _calcResidualVector_(self):
        overlappingVector, overlappingRHSvector = self.globalVectors
        Lx = self.matrix * overlappingVector
        residual = Lx - overlappingRHSvector
        Lx.destroy()
        return residual

    def _calcResidual(self, residualFn=None):
        if residualFn is not None:
            return residualFn(self.var, self.matrix, self.RHSvector)
        else:
            comm = self.var.mesh.communicator
            residual = self._calcResidualVector_()
            norm = comm.Norm2(residual)
            residual.destroy()
            return norm
        
    def _calcRHSNorm(self):
        return self.nonOverlappingRHSvector.Norm2()

    def __del__(self):
        if hasattr(self, "_globalVectors"):
            overlappingVector, overlappingRHSvector = self.globalVectors
            del self.matrix
            overlappingVector.destroy()
            overlappingRHSvector.destroy()




import logging

_log = logging.getLogger(__name__)

import atexit

import pyamgx

from .preconditioners import *

from .pyAMGXSolver import *
from .linearCGSolver import *
from .linearGMRESSolver import *
from .linearFGMRESSolver import *
from .linearBiCGStabSolver import *
from .linearLUSolver import *
from .aggregationAMGSolver import *
from .classicalAMGSolver import *

from . import pyamgxConvergence

pyamgx.initialize()
atexit.register(pyamgx.finalize)

DefaultSolver = LinearCGSolver
DefaultAsymmetricSolver = LinearLUSolver
DummySolver = DefaultSolver
GeneralSolver = LinearLUSolver



import os
import re
import shutil
from subprocess import Popen, PIPE
import sys
import tempfile

from fipy.tools.parser import parse

from examples.benchmarking.utils import monitor

def main():
    url = parse('--svnURL', action='store',
                  type='string', default=None)

    revision = parse('--svnRevision', action='store',
                     type='string', default="HEAD:0")

    args = sys.argv[1:]

    import pysvn

    m = re.match(r"(\d+|HEAD):(\d+|HEAD)", revision)

    def str2rev(s):
        if s == "HEAD":
            return pysvn.Revision(pysvn.opt_revision_kind.head)
        else:
            return pysvn.Revision(pysvn.opt_revision_kind.number, int(s))

    if m is None:
        revisionStart = revisionEnd = str2rev(revision)
    else:
        revisionStart = str2rev(m.group(1))
        revisionEnd = str2rev(m.group(2))

    client = pysvn.Client()

    # svn manipulations on the working copy in-place are dangerous

    if url is None:
        info = client.info('.')
        url = info.url

    dir = tempfile.mkdtemp()

    env = os.environ.copy()
    env['PYTHONPATH'] = dir

    benchmarker = os.path.join(os.path.dirname(__file__),
                               "benchmarker.py")

    pyth = sys.executable or "python"

    try:
        client.checkout(url, dir)

        scanf_e = r"[-+]?(\d+(\.\d*)?|\d*\.\d+)([eE][-+]?\d+)?"

        reCPU = re.compile("cpu time: (%s) s / step / cell" % scanf_e)
        reRSZ = re.compile("max resident memory: (%s) B / cell" % scanf_e)
        reVSZ = re.compile("max virtual memory: (%s) B / cell" % scanf_e)

        for entry in client.log(url,
                                revision_start=revisionStart,
                                revision_end=revisionEnd):
            try:
                client.update(dir, revision=entry.revision)

                p = Popen([pyth, benchmarker] + args
                          + ["--numberOfSteps=0"],
                          stdout=PIPE,
                          stderr=PIPE)

                cpu0, rsz0, vsz0 = monitor(p)

                p = Popen([pyth, benchmarker] + args
                          + ["--cpuBaseLine=%f" % cpu0],
                          stdout=PIPE,
                          stderr=PIPE,
                          env=env)

                cpu, rsz, vsz = monitor(p)

                print(entry.revision.number, cpu, rsz, vsz)

            except Exception as e:
                print(entry.revision.number, e)
    except Exception as e:
        print(e)

    shutil.rmtree(dir)

if __name__ == "__main__":
    main()
